package rione.agent;

import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.messages.Command;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.communication.CommunicationUtil;
import rione.communication.Port;
import rione.communication.CommunicationManager;
import rione.communication.CommunicationManagerPartsBuilder;
import rione.communication.MessageBitSection;
import rione.geom.AreaSplit;
import rione.io.BitArrayInputStream;
import rione.util.BitUtil;

public abstract class CommunicationAgent<E extends StandardEntity> extends Agent<E> {

	private CommunicationManager comManager = null;

	private Map<EntityID, EntityID> aggregator = null;
	private Set<EntityID> messenger = null;

	@Override
	protected void initialize() {
		super.initialize();

		CommunicationManagerPartsBuilder builder = new CommunicationManagerPartsBuilder(me(), world);
		CommunicationUtil comUtil = builder.getCommunicationUtil();

		// TODO comUtil.isNoRadio, isNoVoiceを使って場合分けする
		// 必ず世界大会までのNoCom対策を！

		/*
		 * Radio/Voice系 全てのエージェントで同じ物をaddされるようにすること！
		 */
		// ダミー，通信が全て正常にできているか確認するときに使う
		// final Port dummy = new DummyPort();
		// builder.addRadioEventListener(dummy);
		if (comUtil.isNoRadio()) {
			System.out.print("NoRadio, ");
			decideNoComAgents();

			// 自分の場所
			final Port ownInfoPort = createNoComPort(comUtil);
			builder.addVoiceEventListener(ownInfoPort);
			final Port locateInBlocadePort = world.getLocateInBlockade().createPort(me(), comUtil);
			builder.addRadioEventListener(locateInBlocadePort);
			builder.addVoiceEventListener(locateInBlocadePort);
			// 火災発見伝令Voice
			final Port burningBuildingsPort = world.getBurningBuildings().createNoComPort(comUtil, isAggregator(), isMessenger());
			builder.addRadioEventListener(burningBuildingsPort);
			builder.addVoiceEventListener(burningBuildingsPort);
			final Port burningBuildingsRemovePort = world.getBurningBuildings().createRemovePort(comUtil);
			builder.addRadioEventListener(burningBuildingsRemovePort);
			builder.addVoiceEventListener(burningBuildingsRemovePort);
			// NeedRescueHuman
			final Port buriedHumansAddPort = world.getBuriedHumans().createAddPort(comUtil);
			builder.addVoiceEventListener(buriedHumansAddPort);
			final Port buriedHumansRemovePort = world.getBuriedHumans().createRemovePort(comUtil);
			builder.addVoiceEventListener(buriedHumansRemovePort);
		}
		else {
			// 自分
			final Port ownInfoPort = createPort(comUtil);
			builder.addRadioEventListener(ownInfoPort);
			final Port locateInBlocadePort = world.getLocateInBlockade().createPort(me(), comUtil);
			builder.addRadioEventListener(locateInBlocadePort);
			// CriticalArea
			final Port criticalAreaPort = world.getCriticalArea().createPort(comUtil);
			builder.addRadioEventListener(criticalAreaPort);
			// builder.addVoiceEventListener(criticalAreaCEL);
			// BurningBuildings
			final Port burningBuildingsAddPort = world.getBurningBuildings().createAddPort(comUtil);
			builder.addRadioEventListener(burningBuildingsAddPort);
			builder.addVoiceEventListener(burningBuildingsAddPort);
			final Port burningBuildingsRemovePort = world.getBurningBuildings().createRemovePort(comUtil);
			builder.addRadioEventListener(burningBuildingsRemovePort);
			builder.addVoiceEventListener(burningBuildingsRemovePort);
			// NeedRescueHuman
			final Port buriedHumansAddPort = world.getBuriedHumans().createAddPort(comUtil);
			builder.addRadioEventListener(buriedHumansAddPort);
			builder.addVoiceEventListener(buriedHumansAddPort);
			final Port buriedHumansRemovePort = world.getBuriedHumans().createRemovePort(comUtil);
			builder.addRadioEventListener(buriedHumansRemovePort);
			builder.addVoiceEventListener(buriedHumansRemovePort);
		}

		/*
		 * Civilian系
		 */
		// worldへの追加が最初
		builder.addCivilianVoiceListener(world.createNewCivilianListener());
		builder.addCivilianVoiceListener(world.getBuriedHumans().createCivilianVoiceListener());

		comManager = new CommunicationManager(builder);
	}

	private void decideNoComAgents() {
		Set<EntityID> messengersID = new HashSet<EntityID>();
		List<EntityID> aggregatorList = new ArrayList<EntityID>();
		// AT
		final Collection<StandardEntity> at = world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM);
		final int atSize = at.size();
		if (2 < atSize) {
			for (StandardEntity se : at) {
				final EntityID id = se.getID();
				if (atSize < 5) {
					if (world.getUniform().toUniform(id) == 1) {
						aggregatorList.add(id);
					}
				}
				else {
					if (world.getUniform().toUniform(se.getID()) < atSize / 3) {
						aggregatorList.add(id);
					}
				}
			}
		}
		// PF
		final Collection<StandardEntity> pf = world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE);
		final int pfSize = pf.size();
		if (2 < pfSize) {
			for (StandardEntity se : pf) {
				final EntityID id = se.getID();
				if (pfSize < 5) {
					if (world.getUniform().toUniform(id) == 1) {
						aggregatorList.add(id);
					}
				}
				else {
					int uniform = world.getUniform().toUniform(se.getID());
					if (uniform < pfSize / 4) {
						messengersID.add(id);
					}
					else if (uniform < Math.max(pfSize / 3, pfSize / 4 + 1)) {
						aggregatorList.add(id);
					}
				}
			}
		}
		aggregator = Collections.unmodifiableMap(createAggregators(aggregatorList));
		messenger = Collections.unmodifiableSet(messengersID);
	}

	private Map<EntityID, EntityID> createAggregators(List<EntityID> aggregatorList) {
		final int aggregatorSize = aggregatorList.size();
		final Map<EntityID, EntityID> result = new HashMap<EntityID, EntityID>(aggregatorSize);

		Collections.sort(aggregatorList, new Comparator<EntityID>() {
			@Override
			public int compare(EntityID id1, EntityID id2) {
				return id1.getValue() - id2.getValue();
			}
		});
		Rectangle[] rects = new AreaSplit(world, aggregatorSize).getRects();
		for (int i = 0; i < aggregatorSize; i++) {
			Area near = world.getNearestRoad((int) rects[i].getCenterX(), (int) rects[i].getCenterY());
			assert (near != null);
			result.put(aggregatorList.get(i), near.getID());
		}
		return result;
	}

	private Port createNoComPort(final CommunicationUtil comUtil) {
		return new Port() {
			boolean createdFlg = false;

			@Override
			public boolean hasNext() {
				return !createdFlg && (me() instanceof Human);
			}

			@Override
			public void init(ChangeSet changed) {
				createdFlg = false;
			}

			@Override
			public MessageBitSection next() {
				createdFlg = true;
				MessageBitSection sec = new MessageBitSection(10);
				// Location
				comUtil.writeArea(sec, (Area) location(), world.getUniform());

				return sec;
			}

			@Override
			public void read(final EntityID senderID, final int time, final BitArrayInputStream stream) {
				// 読み
				final StandardEntity sender = world.getEntity(senderID);
				final EntityID areaID = comUtil.readArea(stream, world.getUniform());
				if (areaID == null)
					return;
				final Area area = (Area) world.getEntity(areaID);

				// ノイズで間違っている
				if (sender == null || area == null)
					return;

				if (world.getTimestamp().getLastChangedTime(senderID) < time) {
					if (sender instanceof Human) {
						Human sendHuman = (Human) sender;
						sendHuman.setPosition(area.getID(), area.getX(), area.getY());
					}
				}

			}
		};
	}

	/**
	 * 自分の情報をあつかう
	 * 
	 * @param comUtil
	 * @return
	 */
	public Port createPort(final CommunicationUtil comUtil) {
		final int FB_TANK_BIT_SIZE = BitUtil.needBitSize(world.getConfig().maxTankCapacity / world.getConfig().maxPower);
		return new Port() {
			boolean createdFlg = false;

			@Override
			public boolean hasNext() {
				return !createdFlg && (me() instanceof Human);
			}

			@Override
			public void init(ChangeSet changed) {
				createdFlg = false;
			}

			@Override
			public MessageBitSection next() {
				createdFlg = true;
				MessageBitSection sec = new MessageBitSection(10);
				// Location
				comUtil.writeArea(sec, (Area) location(), world.getUniform());

				if (me() instanceof AmbulanceTeam) {
				}
				else if (me() instanceof FireBrigade) {
					sec.add(((FireBrigade) me()).getWater() / world.getConfig().maxPower, FB_TANK_BIT_SIZE);
				}
				else if (me() instanceof PoliceForce) {
				}
				return sec;
			}

			@Override
			public void read(final EntityID senderID, final int time, final BitArrayInputStream stream) {
				// 読み
				final StandardEntity sender = world.getEntity(senderID);
				final EntityID areaID = comUtil.readArea(stream, world.getUniform());
				Area area = null;
				if (areaID != null) {
					area = (Area) world.getEntity(areaID);
				}
				if (sender instanceof FireBrigade) {
					FireBrigade senderFB = (FireBrigade) sender;
					final int tank = stream.readBit(FB_TANK_BIT_SIZE);
					senderFB.setWater(tank * world.getConfig().maxPower);
				}

				// ノイズで間違っている
				if (sender == null || area == null)
					return;

				if (world.getTimestamp().getLastChangedTime(senderID) < time) {
					if (sender instanceof Human) {
						Human sendHuman = (Human) sender;
						sendHuman.setPosition(area.getID(), area.getX(), area.getY());
					}
				}

			}
		};
	}

	@Override
	protected void hear(Collection<Command> heard) {
		comManager.read(heard);
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();

		comManager.update();
		// チャネルの設定
		sendSubscribe(time, comManager.getSubscribeChannels());
	}

	@Override
	protected void afterAct() {
		super.afterAct();
		int channel = comManager.getSendChannel();
		if (0 <= channel) {
			sendSpeak(time, channel, comManager.createSendMessage(channel, changed));
		}
		if (me() instanceof Human) {
			channel = comManager.getSpeakChannel();
			if (0 <= channel) {
				sendSpeak(time, channel, comManager.createSpeakMessage(channel, changed));
			}
		}
	}

	public boolean isNoRadio() {
		return aggregator != null;
	}

	public boolean isAggregator() {
		return isNoRadio() && aggregator.containsKey(getID());
	}

	public EntityID getAggregatorPosition(EntityID id) {
		return aggregator.get(id);
	}

	public Collection<EntityID> getAggregators() {
		return aggregator.keySet();
	}

	public Collection<EntityID> getMessengers() {
		return messenger;
	}

	public boolean isMessenger() {
		return isNoRadio() && messenger.contains(getID());
	}

}
