package rione.agent;

import java.util.*;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.at.AmbulanceTeamAgent;
import rione.model.AgentConstants;
import rione.model.Complex;
import rione.model.ConfigConstants;
import rione.model.ConfigConstants.VoiceChannel;
import rione.util.DistanceSorter;

public abstract class PlatoonAgent<E extends Human> extends HumanoidAgent<E> {

	final protected Map<Integer, EntityID> locationHistory = new HashMap<Integer, EntityID>();

	/**
	 * 他のエージェントが探索済みのエリア
	 */
	protected Set<EntityID> someoneVisitedArea = new HashSet<EntityID>();

	/**
	 * 確認していないエリア
	 */
	protected Set<EntityID> unenteredBuildings = null;

	protected Set<EntityID> unlookupedBuildings = null;

	protected Set<StandardEntity> unlookupedRoadsInFrontOfTheBuilding = new HashSet<StandardEntity>();

	/**
	 * 埋まっている市民が密集している集合
	 */
	protected Set<Building> PrioritizeBuilding = new HashSet<Building>();

	@Override
	protected void initialize() {
		super.initialize();
		initUnentered();
		initUnlookuped();
	}

	private void initUnlookuped() {
		Collection<StandardEntity> initedBuildings = world.getEntitiesOfType(StandardEntityURN.BUILDING);
		unlookupedBuildings = new HashSet<EntityID>(initedBuildings.size());
		for (StandardEntity se : initedBuildings) {
			unlookupedBuildings.add(se.getID());
		}
	}

	private void initUnentered() {
		StandardEntityURN[] initedURN;
		if (this instanceof AmbulanceTeamAgent) {
			initedURN = new StandardEntityURN[] {
					StandardEntityURN.BUILDING,
					StandardEntityURN.AMBULANCE_CENTRE,
					StandardEntityURN.FIRE_STATION,
					StandardEntityURN.POLICE_OFFICE,
			};
		}
		else {
			initedURN = new StandardEntityURN[] {
					StandardEntityURN.BUILDING,
			};
		}
		Collection<StandardEntity> initedBuildings = world.getEntitiesOfType(initedURN);
		unenteredBuildings = new HashSet<EntityID>(initedBuildings.size());
		for (StandardEntity se : initedBuildings) {
			unenteredBuildings.add(se.getID());
		}
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();
		locationHistory.put(time, location().getID());

		for (StandardEntity se : world.getEntitiesOfType(AgentConstants.PLATOONS)) {
			Human human = (Human) se;
			if (human.isPositionDefined()) {
				someoneVisitedArea.add(human.getPosition());
			}
		}
		passedRoad();
		unlookupedBuildings.removeAll(getVisibleEntities());
	}

	/**
	 * PlatoonAgent共通して行うべき行動
	 */
	@Override
	protected void act() throws ActionCommandException {
		// 建物から出られないときの対策
		if(getCommandHistory(time-1) != null && getCommandHistory(time-2) != null){
			if (getCommandHistory(time - 1).equals(StandardMessageURN.AK_MOVE)
				&& getCommandHistory(time - 2).equals(StandardMessageURN.AK_MOVE)){
				if (locationHistory.get(time - 1).equals(locationHistory.get(time - 2))) {
					if (time % 4 == 0 && location() instanceof Building) {
						Building bld = (Building) location();
						goOut(bld);
					}
					if(time % 3 == 0 && location() instanceof Road){
						randomWalk();
					}
				}
			}
		}
	}

	public void MoveToError() throws ActionCommandException{
		if (locationHistory.get(time - 1).equals(locationHistory.get(time - 2)) &&
				locationHistory.get(time - 2).equals(locationHistory.get(time - 3)) &&
				locationHistory.get(time - 3).equals(locationHistory.get(time - 4)) &&
				time == 10){
			move(location(),me().getX()+1000,me().getY()+1000);
		}
	}


	private Iterator<Road> entranceIt = null;

	/**
	 * 建物の入り口に移動する。
	 *
	 * @param bld
	 * @throws ActionCommandException
	 */
	protected void goOut(Building bld) throws ActionCommandException {
		// 中央から離れている場合は一旦中央へ
		if (1000 < world.getDistance(location(), me())) {
			move(location(), location().getX(), location().getY());
		}
		if (entranceIt == null || !entranceIt.hasNext()) {
			entranceIt = world.getEntrance().getEntrance(bld).iterator();
		}
		Road entranceRoad = entranceIt.next();
		move(entranceRoad);
	}

	/**
	 * Buildingを外から見てまわる
	 *
	 * @throws ActionCommandException
	 */
	protected void lookupSearchBuildings() throws ActionCommandException {
		if (unlookupedBuildings.isEmpty()) {
			initUnlookuped();
		}
		// 誰かが到達したことのあるBuildingを除く
		unlookupedBuildings.removeAll(someoneVisitedArea);
		// 燃えた建物を除外
		for (Iterator<EntityID> it = unlookupedBuildings.iterator(); it.hasNext();) {
			EntityID id = it.next();
			StandardEntity se = world.getEntity(id);
			if (se instanceof Building) {
				Building building = (Building) se;
				if (building.isOnFire()
						|| (building.isFierynessDefined() && building.getFierynessEnum() == Fieryness.BURNT_OUT)) {
					it.remove();
				}
			}
		}
		if (!unlookupedBuildings.isEmpty()) {
			Set<StandardEntity> dest = new HashSet<StandardEntity>(unlookupedBuildings.size());
			for (EntityID id : unlookupedBuildings) {
				dest.add(world.getEntity(id));
			}
			if (me() instanceof PoliceForce) {
				moveFront(dest, router.getPfCostFunction());
			}
			moveFront(dest);
		}
	}

	/**
	 * Buildingの中に入って見まわり
	 *
	 * @throws ActionCommandException
	 */
	protected void enterSearchBuildings() throws ActionCommandException {
		if (unenteredBuildings.isEmpty()) {
			initUnentered();
		}
		// 誰かが到達したことのあるBuildingを除く
		unenteredBuildings.removeAll(someoneVisitedArea);
		// 燃えた建物を除外
		for (Iterator<EntityID> it = unenteredBuildings.iterator(); it.hasNext();) {
			EntityID id = it.next();
			StandardEntity se = world.getEntity(id);
			if (se instanceof Building) {
				Building building = (Building) se;
				if (building.isOnFire()
						|| (building.isFierynessDefined() && building.getFierynessEnum() == Fieryness.BURNT_OUT)) {
					it.remove();
				}
			}
		}
		if (!unenteredBuildings.isEmpty()) {
			Set<StandardEntity> dest = new HashSet<StandardEntity>(unenteredBuildings.size());
			for (EntityID id : unenteredBuildings) {
				if(router.isReachable(id))
					dest.add(world.getEntity(id));
			}
			if(dest.isEmpty()) return;
			if (me() instanceof PoliceForce) {
				move(dest, router.getPfCostFunction());
			}else{
				move(dest);
			}
		}
	}

	/**
	 * Buildingの前の道を見まわり
	 *
	 * @throws ActionCommandException
	 */
	protected void lookupSearchRoads() throws ActionCommandException {
		int min = Integer.MAX_VALUE;
		StandardEntity temp = null;
		StandardEntity pos = world.getEntity(me().getPosition());
		for(StandardEntity road : unlookupedRoadsInFrontOfTheBuilding){
			int distance = world.getDistance(pos,road);
			Area area = (Area)road;
			if(router.isSureReachable(area) && distance < min){
				min = distance;
				temp = road;
			}
		}
		if(temp != null){
			move(temp);
		}
	}


	//Areaに接続れさた建物を返す
		protected Collection<Building> getneighBuildings(Area area){
			Collection<Building> building =new LinkedHashSet<Building>();
			for(EntityID ent:area.getNeighbours()){
				Area sEnt=(Area)world.getEntity(ent);
				if(checkEntrance(sEnt)){
					for(EntityID build:sEnt.getNeighbours()){
						if(world.getEntity(build) instanceof Building ){
							building.add((Building) world.getEntity(build));
						}
					}
				}else if(sEnt instanceof Building){
					building.add((Building)sEnt);
				}
			}
			return building;
		}

		//AreaがEntranceかどうか判定
		protected boolean checkEntrance(Area area){
			for(EntityID ent:area.getNeighbours()){
				if(world.getEntity(ent) instanceof Building )return true;
			}
			return false;
		}

		protected Set<Area> getThreeNeighbourRoad(Area area){
			Set<Area> neighborSet = new LinkedHashSet<Area>();
			for(EntityID oneStepId : area.getNeighbours()){
				StandardEntity oneStepSe = world.getEntity(oneStepId);
				if(!(oneStepSe instanceof Road))continue;
				neighborSet.add((Area)oneStepSe);
				for(EntityID twoStepId:((Area)oneStepSe).getNeighbours()){
					StandardEntity twoStepSe = world.getEntity(twoStepId);
					if(!(twoStepSe instanceof Road))continue;
					neighborSet.add((Area)twoStepSe);
				}
			}
			return neighborSet;
		}

		/**
		 * Roadと距離(Roadの数)を入れると､距離の間にあるBuildingの集合を返す
		 * @param startRoad
		 * @param maxDistance
		 * @return maxDistance個のRoadを挟んだ間にあるBuildngの集合
		 */
		protected Collection<Building> getNeighbourBuilding(Road startRoad, int maxDistance){
			Collection<StandardEntity> aroundEntity = new HashSet<StandardEntity>(world.getObjectsInRange(startRoad.getID(), 100000));
			Collection<EntityID> buildingEntrance = new HashSet<EntityID>();
			Collection<Building> neighbourBuilding = new HashSet<Building>();

			Collection<EntityID> confirmedRoad = new HashSet<EntityID>();
			HashMap<EntityID, Building> entranceToBuilding = new HashMap<EntityID, Building>();
			for(StandardEntity se : aroundEntity){
				if(se instanceof Building){
					buildingEntrance.addAll(((Building) se).getNeighbours());
					for(EntityID entrance : ((Building) se).getNeighbours()){
						if(world.getEntity(entrance) instanceof Road){
							entranceToBuilding.put(entrance, (Building)se);
						}
					}
				}
			}
			LinkedHashSet<EntityID> ll = new LinkedHashSet<EntityID>();
			ll.add(startRoad.getID());
			confirmedRoad.add(startRoad.getID());

			int count = 0;
			int nowCycleSize = 0;
			int nextCycleSize = startRoad.getNeighbours().size();

			while(!buildingEntrance.isEmpty() && count < maxDistance && ll.size() != 0){
				Road r = (Road)world.getEntity(ll.iterator().next());

				for(EntityID neighbour : r.getNeighbours()){
					if(world.getEntity(neighbour) instanceof Road){
						ll.add(neighbour);
					}
				}
				int diff = ll.size();
				ll.removeAll(confirmedRoad);
				diff -= ll.size();
				confirmedRoad.add(r.getID());
				if(buildingEntrance.contains(r.getID())){
					neighbourBuilding.add(entranceToBuilding.get(r.getID()));
					buildingEntrance.remove(r.getID());
				}

				//HACK: startRoadからの距離を取るために以下の方法でカウントしている｡
				if(nowCycleSize == 0){
					nowCycleSize = nextCycleSize;
					nextCycleSize = 0;
					count++;
				}
				else {
					nextCycleSize += r.getNeighbours().size()-diff;
					nowCycleSize -= 1;
				}
			}

			return neighbourBuilding;
		}

	/**
	 * unlookupedRoadsInFrontOfTheBuildingに到達したかどうか　　　　　　　
	 * initUnlookupedRoadsで集合を作成
	 *
	 */
	protected void passedRoad() {
		if(unlookupedRoadsInFrontOfTheBuilding.isEmpty()){
			initUnlookupedRoads();
		}
		if(unlookupedRoadsInFrontOfTheBuilding.isEmpty())return;
		StandardEntity pos = world.getEntity(me().getPosition());
		Set<StandardEntity> temp = new HashSet<StandardEntity>();
		for(StandardEntity road : unlookupedRoadsInFrontOfTheBuilding){
			Area area = (Area)road;
			if(pos.getID().equals(road.getID())){
				temp.add(road);
			}else if(isVisible(road) && router.isSureReachable(area)){
				temp.add(road);
			}
		}
		if(!temp.isEmpty()){
			unlookupedRoadsInFrontOfTheBuilding.removeAll(temp);
		}
	}

	/**
	 * 建物前の道の集合を作成
	 */
	protected void initUnlookupedRoads(){
		Set<StandardEntity> lookingTemp = new HashSet<StandardEntity>();
		for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.BUILDING)){
			for(StandardEntity area : world.getNeighbours(se, 10)){
				if(area instanceof Road){
					lookingTemp.add(area);
				}
			}
		}
		for(StandardEntity se : lookingTemp){
			StandardEntity temp = null;
			int min = Integer.MAX_VALUE;
			for(StandardEntity area : world.getNeighbours(se, 10)){
				int distance = world.getDistance(se, area);
				if(distance < min && se.getID() != area.getID() && area instanceof Road){
					min = distance;
					temp = area;
				}
			}
			if(temp != null){
				unlookupedRoadsInFrontOfTheBuilding.add(temp);
			}
		}
	}
	/**
	 * RandomWalk
	 * @throws ActionCommandException
	 */
	protected void randomWalk() throws ActionCommandException {
		ArrayList<StandardEntity> areas = new ArrayList<StandardEntity>(world.getEntitiesOfType(AgentConstants.AREAS));
		areas.removeAll(world.getIsolated());
		Area area = (Area) areas.get(random.nextInt(areas.size()));
		move(area);
	}

	/**
	 * 近くの避難所に移動する
	 *
	 * @throws ActionCommandException
	 */
	protected void moveToRefuge() throws ActionCommandException {
		int mindistance = Integer.MAX_VALUE;
		StandardEntity tempRefuge = null;
		StandardEntity agent = world.getEntity(me().getPosition());
		Area pos = null;
		if(agent instanceof Area){
			pos = (Area)agent;
		}
		for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.REFUGE)){
			if(se instanceof Refuge){
				int distance = world.getDistance(se,location());
				if(mindistance > distance){
					mindistance = distance;
					tempRefuge = se;
				}
			}
		}
		if(tempRefuge!=null){
			//move(tempRefuge);
			if (tempRefuge instanceof Area) {
				move((Area)tempRefuge, pos);
			}
		}
	}

	/**
	 * aggregatorPosition周辺をうろうろする
	 * @throws ActionCommandException
	 */
	protected void aggregatorStay() throws ActionCommandException {
		EntityID agPos = getAggregatorPosition(getID());
		int dist = world.getDistance(me().getPosition(), agPos);
		if (dist < Collections.min(world.getConfig().voiceChannels.values(),
				new Comparator<ConfigConstants.VoiceChannel>() {
			@Override
			public int compare(VoiceChannel c1, VoiceChannel c2) {
				return c1.range - c2.range;
			}
		}).range) {
			randomWalk();
		}
		move(agPos);
	}

	/**
	 * aggregatorを古い順に回る
	 * @param interval
	 * @throws ActionCommandException
	 */
	protected void messengerLoop(int interval) throws ActionCommandException {
		int lastTime = -1;
		EntityID lastSeenID = null;
		for (EntityID id : getAggregators()) {
			int t = world.getTimestamp().getLastSeenTime(getAggregatorPosition(id));
			if (lastTime < t) {
				lastTime = t;
				lastSeenID = id;
			}
		}
		if (lastTime == -1) {
			Set<Area> aggregatorsArea = new HashSet<Area>(getAggregators().size());
			for (EntityID id : getAggregators()) {
				aggregatorsArea.add((Area) world.getEntity(getAggregatorPosition(id)));
			}
			move(aggregatorsArea);
		}
		if (!world.getBurningBuildings().isEmpty() || lastTime - time > interval) {
			move(lastSeenID);
		}
	}

	protected void careSelf() throws ActionCommandException {
		if (me().getHP() - me().getDamage() * (timestep - time) < 16) {
			moveToRefuge();
		}
	}

	/**
	 * 密度を計算する
	 * @throws ActionCommandException
	 */
	protected void prioritize() throws ActionCommandException {
		PrioritizeBuilding.clear();
		double diameter = 40000.0;
		final long t = System.currentTimeMillis();
		Complex[] p = new Complex[1000];

		int count = 0;
		for(EntityID id : world.getBuriedHumans()){
			Human human = (Human) world.getEntity(id);
			if (human == null || !human.isPositionDefined())continue;
			StandardEntity pos = world.getEntity(human.getPosition());
			if (!(pos instanceof Building))continue;
			Building building = (Building) pos;
			Complex temp = new Complex(building.getX(), building.getY());
			p[count] = temp;
			count++;
		}

		int max = 1;
		// 2点間の距離
		for (int i = 0; i < count-1; i++) {
			for (int j = i+1; j < count; j++) {
				int sum = 0;
				Set<Building> circleTemp = new HashSet<Building>();
				// ベクトルvを作成
				Complex v1 = Complex.sub(p[j], p[i]);
				// 2点間の距離
				double d = v1.abs();
				if (d < diameter) {
					Complex center = Complex.getCenter(p[i], p[j],diameter / 2);
					for (EntityID id : world.getBuriedHumans()) {
						//時間が掛かりすぎていないか
						if (System.currentTimeMillis() - t > world.getConfig().thinkTime / 2) {
							System.out.println("off: Prioritize");
							return;
						}
						Human human = (Human) world.getEntity(id);
						if (human == null || !human.isPositionDefined()
								|| !human.isHPDefined() || !human.isBuriednessDefined())continue;
						if(human.getHP() == 0 || human.getBuriedness() == 0)continue;
						StandardEntity pos = world.getEntity(human.getPosition());
						if (!(pos instanceof Building))continue;
						Building building = (Building) pos;
						Complex target = new Complex(building.getX(), building.getY());
						if(Complex.abs(target,center) < diameter / 2){
							sum++;
							circleTemp.add(building);
						}
					}
					if (sum > max) {
						//max = sum;
						PrioritizeBuilding.addAll(circleTemp);
					}
				}
			}
		}
	}
	/**
	 * 視界内の温度の高い建物を返す<br>
	 * ただし、視界内に燃えている建物があった場合はnullを返す
	 * @return 視界内で最も温度の高い建物
	 */
	protected Building getHotBuilding(){
		float maxAreaSize=0;
		int maxTemp = 0;
		double minDistance=Integer.MAX_VALUE;
		Building hotBuilding=null;
		//視界内の一番近い高温の建物を探す
		//TODO 視界内でいいのか
		for(EntityID eid : getVisibleEntities()){
			StandardEntity se = world.getEntity(eid);
			if(se instanceof Building){
				Building bld = (Building) se;
				int temp = bld.getTemperature();
				float areaSize = bld.getGroundArea();
				float distance = world.getDistance(me(),bld);
				if(bld.getFierynessEnum() == Fieryness.UNBURNT
						&& temp!=0 && (maxTemp<temp 
						|| (maxTemp==temp && maxAreaSize < areaSize)
						|| (maxAreaSize==areaSize && distance < minDistance) )){
					hotBuilding=(Building)se;
					maxTemp=temp;
					minDistance=world.getDistance(me(),bld);
					maxAreaSize=areaSize;
				}else if(bld.getFierynessEnum() == Fieryness.BURNING){
					return null;
				}
			}
		}
		return hotBuilding;
	}
	
	/**
	 * Entityを中心として指定した数の建物をEntityからの距離順でとってくる
	 * 
	 * @param centerEnt（中心となるStandardEntity）
	 * @param size（取ってくる建物の個数）
	 * @return Buildings
	 */
	protected List<StandardEntity> getBuidingExtent(StandardEntity centerEnt,int size){
		List<StandardEntity> Buildings = new LinkedList<StandardEntity>(world.getObjectsInRange(centerEnt, 50000));
		Buildings.retainAll(world.getEntitiesOfType(StandardEntityURN.BUILDING));
		//距離順にソート
		Collections.sort(Buildings,new DistanceSorter((StandardEntity)centerEnt, world));
		List<StandardEntity> result = new LinkedList<StandardEntity>();
		for(int i=0; i < Math.min(size, Buildings.size()); i++){
			result.add(Buildings.get(i));
		}
		return result;
	}

	/**
	 * ダメージを負ったHumanが何サイクル放置できるか求める
	 * @param human,ATの速さ
	 * @return humanを放置できるサイクル
	 */
	//避難所に運ぶまでのダメージを見てそれ未満にならないように放置(サイクル数を返す)
	//CVの今のHPー{(CVのダメージ＊避難所までのサイクル数)＋(CVのダメージ＊放置できるサイクル数)}　＞　残したいHP
	//放置できるサイクル　＜　(今のHPー残したいHP(100くらい))/CVのダメージー避難所までのサイクル
	/*public double Triage(Human human) {
		Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		if (refuges.isEmpty()) return 0;
		Refuge refuge = (Refuge) Collections.min(refuges, new DistanceSorter(me(), world));

		if (human.isDamageDefined() && human.isHPDefined()) {
			if (human.getDamage() == 0) return 390;//ダメージがない場合、十分な(放置可能)サイクルを返す
			return ((human.getHP() - 100) / human.getDamage() - (world.getDistance(refuge,me())/getDistAverage()));
		}
		return 0;//判定できない場合、放置可能サイクルを0で返す
	}*/

}
