package rione.agent.at;

import java.util.*;

import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardPropertyURN;
import rescuecore2.worldmodel.EntityID;
import rione.util.DistanceSorter;
import rione.util.IdSorter;
import rione.util.IntPropertySorter;

public class AmbulanceTeamAgent extends AbstractAmbulanceTeamAgent {

	@Override
	protected void initialize() {
		super.initialize();
		System.out.println(toString() + " was connected. [id=" + getID() + ", uniform=" + getUniform() + "]");
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();
	}
	
	protected EnumSet<StandardEntityConstants.Fieryness> saveFieryness = EnumSet.of(
			Fieryness.UNBURNT,
			Fieryness.WATER_DAMAGE,
			Fieryness.MINOR_DAMAGE);

	@Override
	protected void act() throws ActionCommandException {
		super.act();
		carry();
		careSelf();
		//TODO Aggregatorではなく，Roadで視界に燃えていないBuildingが無ければ燃えていないBuildingへ
		loadOrRescueAround();
		if (30 < time && isAggregator()) {
			aggregatorStay();
		}
		if (30 < time && isMessenger()) {
			messengerLoop(25);
		}
		//prioritizeMove();
		matchingATandVictim();
		prioritizeMove();
		moveToTarget();
		searchAroundHotBuilding();
		lookupSearchRoads();
		randomWalk();
	}

	/**
	 * prioritizeで市民の最も集中している場所に向かう
	 * @throws ActionCommandException
	 */
	protected void prioritizeMove()throws ActionCommandException {
		prioritize();
		if(PrioritizeBuilding.isEmpty() || world.getBuriedHumans() == null)return;
		int minDistance = Integer.MAX_VALUE;
		Building temp = null;
		//自分とPrioritizeBuildingの集合の中で最も距離が近い建物を調べる
		for (Building building : PrioritizeBuilding) {
			int distance = world.getDistance(building, me());
			if(router.isSureReachable(building) && distance < minDistance 
					&& distance < getDistAverage() * 10){
				minDistance = distance;
				temp = building;
			}
		}
		if(temp==null)return;
		StandardEntity pos = world.getEntity(me().getPosition());
		if(pos.getID().equals(temp.getID())){
			PrioritizeBuilding.remove(temp);
		}
		move(temp);
	}
	
	/**
	 * 視界内のHumanをloadできる(背負える)ならloadする． RescueできるならRescueする．
	 * 複数人いる場合、自分の担当を判断する。
	 * 
	 * @throws ActionCommandException
	 */
	protected void loadOrRescueAround() throws ActionCommandException {
		// 視界内のHumanの諦め判定
		if (location() instanceof Building) {
			Building building = (Building) location();
			if (building.isFierynessDefined()
				&& !saveFieryness.contains(building.getFierynessEnum())){
				return;
			}
		}
		
		// 視界内のATとHumanを割り当てる
		List<Civilian> cvLoadList = new ArrayList<Civilian>();
		List<Human> humanRescueList = new ArrayList<Human>();
		List<AmbulanceTeam> atList = new ArrayList<AmbulanceTeam>();
		for (EntityID id : getVisibleEntities()) {
			StandardEntity se = world.getEntity(id);
			if (se instanceof Human) {
				Human hum = (Human) se;
				// (Refuge,Road,ATの上の人を)除外
				if (hum.isPositionDefined()) {
					if (world.getEntity(hum.getPosition()) instanceof Refuge
							|| world.getEntity(hum.getPosition()) instanceof Road
							|| world.getEntity(hum.getPosition()) instanceof AmbulanceTeam) {
						continue;
					}
					//同じ場所にいない人を除外
					if (!world.getEntity(hum.getPosition()).equals(location())) {
						continue;
					}
				}
				//体力0の人を除外
				if (hum.isHPDefined() && hum.getHP() == 0) {
					continue;
				}
				// 救助を行う人
				if (se instanceof AmbulanceTeam && hum.isBuriednessDefined() && hum.getBuriedness() == 0) {
					atList.add((AmbulanceTeam) se);
					continue;
				}
				// 救助される人
				if (hum.isBuriednessDefined()) {
					if (hum.getBuriedness() == 0) {
						if (hum instanceof Civilian) {
							cvLoadList.add((Civilian) hum);
						}
					}
					else {
						humanRescueList.add(hum);
					}
				}
			}
		}
		if (cvLoadList.size() + humanRescueList.size() == 0)
			return;
		// ソート
		// 同じ建物で同じように見えていれば一意に決定される．
		// TODO Distanceの方が通信の情報も使えてていいかも(ただし複雑)
		Collections.sort(cvLoadList, new IdSorter());
		Collections.sort(humanRescueList, new IntPropertySorter<Human>(StandardPropertyURN.BURIEDNESS));
		Collections.sort(atList, new IdSorter());

		// atListのインデックス
		int i = 0;
		boolean judge = false;
		try {
			// 運べる市民がいても近くに市民がいるなら救助を優先する
			/*for (EntityID id : getVisibleEntities()) {
				StandardEntity se = world.getEntity(id);
				if (se instanceof Civilian) {
					Civilian cv = (Civilian) se;
					if (cv.isBuriednessDefined() && cv.getBuriedness() > 0){
						judge = true;
					}
					double remainingTime = triage(cv);
					if(remainingTime < 16){
						judge = false;
						break;
					}	
				}
			}*/
			// とりあえず運べるやつは運ぶ
			if (judge == false) {
				for (; i < cvLoadList.size(); i++) {
					if (atList.get(i).equals(me())) {
						load(cvLoadList.get(i));
					}
				}
			}	
			// 早く助かる人間から助ける
			for (Human hum : humanRescueList) {
				// 実験により一人のrescueで1だけburiednessが減る模様．
				// それらしきconfigもないので決め打ち．
				for (int j = 0; j < hum.getBuriedness(); j++) {
					if (atList.get(i).equals(me())) {
						rescue(hum);
					}
					i++;
				}
			}
		}
		catch (IndexOutOfBoundsException e) {
			// 範囲外のインデックスを指定したら終了．
			// ・Loadすべき人数よりATが少ない．
			// ・Rescueに必要な人数よりATが少ない．
		}
	}


	/**
	 * ターゲットに移動する
	 * 
	 * @throws ActionCommandException
	 */
	protected void moveToTarget() throws ActionCommandException {
		Set<Area> targetSureArea = new HashSet<Area>();
		Set<Area> targetArea = new HashSet<Area>();

		// 1[km]以内のAT
		int nearAT = 0;
		for (StandardEntity se : world.getObjectsInRange(me(), 1000 * 1000)) {
			if (se instanceof AmbulanceTeam) {
				AmbulanceTeam at = (AmbulanceTeam) se;
				if (at.isBuriednessDefined() && at.getBuriedness() > 0)
					continue;
				nearAT++;
			}
		}
		if (nearAT == 0) {
			nearAT = 1;
		}

		//TODO 判断基準がworld.needRescueHumans.updateと二重になっていておかしい
		//TODO 近くが燃えている場合をどうするか
		
		Map<EntityID, Double> minStaticCost = new HashMap<EntityID, Double>();
		
		// needRescueHumanから，助かりそうな人がいるAreaだけを選ぶ
		for (EntityID id : world.getBuriedHumans()) {
			Human human = (Human) world.getEntity(id);
			if (human == null
					|| !human.isPositionDefined()) {
				continue;
			}
			StandardEntity pos = world.getEntity(human.getPosition());
			if (!(pos instanceof Building)) {
				continue;
			}
			Building building = (Building) pos;
			if (building.equals(location())
					|| building.getID().equals(locationHistory.get(time - 1))
					|| building.getID().equals(locationHistory.get(time - 2))) {
				continue;
			}
			int buriedness;
			if (human.isBuriednessDefined()) {
				buriedness = human.getBuriedness();
			}
			else {
				buriedness = 10;
			}
			
			// 助かる可能性がないHumanは除外
			int deathTime = world.getDeathTime(human);
			//TODO 8は通常のconfigなら最大10ぐらい誤差があるはずなのでこれぐらいかなぁと．
			if (deathTime < buriedness / nearAT + time + 8) {
				continue;
			}
						
			double staticCost = buriedness * getDistAverage();
			//TODO Refugeとの最小距離
			if (minStaticCost.containsKey(building.getID())) {
				minStaticCost.put(building.getID(),
						Math.min(minStaticCost.get(building.getID()), staticCost));
			}
			else {
				minStaticCost.put(building.getID(), staticCost);
			}

			// 到達のみを受け付ける
			if (router.isSureReachable(building)) {
				targetSureArea.add(building);
			}
			int distance = world.getDistance(location(),building);
			if(distance < getDistAverage() * 16 && building.isFierynessDefined() 
					&& !building.getFierynessEnum().equals(Fieryness.BURNT_OUT)
					&& !building.getFierynessEnum().equals(Fieryness.INFERNO)){
				targetArea.add(building);
			}
		}
		if (targetSureArea.size() == 1) {
			// サイズ1ならA*move
			for (Area se : targetSureArea) {
				move(se);
			}
		}
		else if (targetSureArea.size() > 1) {
			// サイズ2以上ならダイクストラmove
			move(targetSureArea, router.getAtCostFunction(minStaticCost));
		}
		if (targetArea.size() == 1) {
			// サイズ1ならA*move
			for (Area se : targetArea) {
				move(se);
			}
		}
		else if (targetArea.size() > 1) {
			// サイズ2以上ならダイクストラmove
			move(targetArea, router.getAtCostFunction(minStaticCost));
		}
	}

	/**
	 * Refugeへ運ぶ(体力0のHumanは降ろす)
	 * 
	 * @throws ActionCommandException
	 */
	protected void carry() throws ActionCommandException {
		Civilian board = someoneOnBoard();
		if (board != null) {
			if (location() instanceof Refuge || (board.isHPDefined() && board.getHP() == 0)) {
				unload();
			}
			else {
				moveToRefuge();
			}
		}
	}

	protected void matchingATandVictim() throws ActionCommandException {
		//助けるVictimの優先順位の決定
		boolean complete;
		ArrayList<Integer> routingAT = new ArrayList<Integer> ();
		double data[][] = new double[100+1][500+1];
		ArrayList<Double> prioritizePoints = new ArrayList<Double>();
		ArrayList<StandardEntity> buriedVictim = new ArrayList<StandardEntity>();
		ArrayList<StandardEntity> activeAT = new ArrayList<StandardEntity>(world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM));
		
		//建物の中で埋もれてる人の集合をとる
		for(EntityID bh : world.getBuriedHumans()){
			//if(router.isSureReachable(bh)){
					buriedVictim.add(world.getEntity(bh));
			//}
		}
		
		//埋まってないATだけに絞る
//		for(Iterator<StandardEntity> iterator = activeAT.iterator(); iterator.hasNext();){			
//			StandardEntity at = iterator.next();
//			if(world.getBuriedHumans().contains(at.getID())){
//				iterator.remove();
//			}
//		}
		for(int i=0;i< activeAT.size(); i++){
			if(world.getBuriedHumans().contains(activeAT.get(i).getID())){
				activeAT.remove(i);
				i--;
			}
		}
		
		//System.out.println(AT.size());
		//System.out.println(buriedVictim.size());
		//System.out.println();
		//Vector<Vector<Double>> data = new Vector<Vector<Double>>(Vic.size(),AT.size());
		//Vector<Double> tmp = new Vector<Double>(AT.size());
		
		for(int i=0; i < activeAT.size(); i++){
			for(int j=0; j < buriedVictim.size(); j++){
				data[i][j] = -1.0;
			}
		}
//		for(int i = 0; i< AT.size(); i++){
//			for (int k = 0; k < Vic.size(); k++) {
//				tmp.add(-1.0);
//			}
//		}
		if(buriedVictim.size() != 0 && activeAT.size() != 0){
		for (int i = 0; i < activeAT.size(); i++) {
				prioritizePoints.add((double)-1);
				for (int j = 0; j < buriedVictim.size(); j++) {
					if(triage((Human)buriedVictim.get(j))==0) data[i][j]=-1;
					else{
						//data[i][j]= 1/triage((Human)Vic.get(j))/Math.pow(world.getDistance(AT.get(i).getID(),Vic.get(j).getID()), 2);
						data[i][j]= 1/triage((Human)buriedVictim.get(j));
						//if(fireDistance((Human)buriedVictim.get(j))!) continue;
						if(fireDistance((Human)buriedVictim.get(j))!=0){//&&
						//((Human) activeAT.get(i)).getHP()>5000){
							Human human = (Human) activeAT.get(i);
							if (human.isHPDefined()) {
								//System.out.println(human.getHP());
								data[i][j]+=10/fireDistance((Human)buriedVictim.get(j));
								
							}
						}
					}
				}
			}
			
			int loop =0;
			do{
				complete=true;
				for(int i=0;i<activeAT.size();i++){
					for(int j=0;j<buriedVictim.size();j++){
						if(data[i][j]>prioritizePoints.get(i)){
							prioritizePoints.add(i, data[i][j]);
							routingAT.add(i, j);
						}else{
							prioritizePoints.add(i, (double)-1);
							routingAT.add(i, -1);
						}
					}
				}
				ArrayList<Integer> routingAT2 = new ArrayList<Integer>();
				routingAT2.addAll(routingAT);
				Collections.sort(routingAT2);
				for(int i=1; i < routingAT2.size(); i++){
					if(prioritizePoints.get(i).intValue()==-1) continue;
					if(routingAT2.get(i) == -1) continue;
					if(routingAT2.get(i) == routingAT2.get(i-1)){
						if(prioritizePoints.get(routingAT.indexOf(routingAT2.get(i))).intValue()
							> prioritizePoints.get(routingAT.lastIndexOf(routingAT2.get(i))).intValue()){
							prioritizePoints.add(routingAT.lastIndexOf(routingAT2.get(i)), (double)-1);
							routingAT.add(routingAT.lastIndexOf(routingAT2.get(i)), -1);
							complete = false;
						}
						else{
							prioritizePoints.add(routingAT.indexOf(routingAT2.get(i)), (double)-1);
							routingAT.add(routingAT.indexOf(routingAT2.get(i)), -1);
							complete = false;
						}
					}
				}
				routingAT2.clear();
				for(int i=0; i<activeAT.size(); i++){
					if(prioritizePoints.get(i).intValue()==-1) continue;
					for(int j=0; j< buriedVictim.size() || j< activeAT.size(); j++){
						if(j<activeAT.size()) data[j][routingAT.get(i)]=-1;
						if(j<buriedVictim.size()) data[i][j]=-1;
					}
				}
				loop++;
				if(loop>=3) break;
			}while(!complete);
			if(!routingAT.isEmpty()){
				if(activeAT.indexOf(world.getEntity(me().getID())) != -1){
					if(routingAT.get(activeAT.indexOf(world.getEntity(me().getID()))) != -1){
						if(buriedVictim.get(routingAT.get(activeAT.indexOf(world.getEntity(me().getID()))))!=null){
							//System.out.println("hoge");
							move(buriedVictim.get(routingAT.get(activeAT.indexOf(world.getEntity(me().getID())))));
							//move(Vic.get(r.get()));
						}
					}
				}
			}
			/*for(int i=0;i<AT.size();i++){
				if(me().getID()==AT.get(i).getID()){
					if(i < r.size()){
						if(r.get(i) != null && r.get(i) != -1){
							move(Vic.get(r.get(i)));
						}
					}
				}
			}*/
		}
		//activeAT.clear();
		//buriedVictim.clear();
		//prioritizePoints.clear();
		//routingAT.clear();
	}
	/**
	 * ダメージを負ったHumanが何サイクル放置できるか求める
	 * @param human,ATの速さ
	 * @return humanを放置できるサイクル
	 */
	//避難所に運ぶまでのダメージを見てそれ未満にならないように放置(サイクル数を返す)
	//CVの今のHPー{(CVのダメージ＊避難所までのサイクル数)＋(CVのダメージ＊放置できるサイクル数)}　＞　残したいHP
	//放置できるサイクル　＜　(今のHPー残したいHP(100くらい))/CVのダメージー避難所までのサイクル
	public double triage(Human human) {
		Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		if (refuges.isEmpty()) return 0;
		Refuge refuge = (Refuge) Collections.min(refuges, new DistanceSorter(me(), world));
	
		if (human.isDamageDefined() && human.isHPDefined()) {
			if (human.getDamage() == 0) return 390;//ダメージがない場合、十分な(放置可能)サイクルを返す
			//return ((human.getHP() - 100) / human.getDamage() - (world.getDistance(refuge,me())/getDistAverage()));
			//System.out.println(human.getHP());
			return ((human.getHP() - 100) / human.getDamage() - (Math.abs(me().getX()-refuge.getX())+Math.abs(me().getY()-refuge.getY()))) / getDistAverage();
		}
		return 0;//判定できない場合、放置可能サイクルを0で返す
	}
	
	public double fireDistance(Human human){
		Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		int min=Integer.MAX_VALUE;
		if (refuges.isEmpty()) return 0;
		for(Building i:world.getBurningBuildings()){
			if(min>world.getDistance(human, i)) min=world.getDistance(human, i);
		}
		if(min>50/Math.sqrt(world.getTime())+10) return min;
		else return 0;
	}
	/**
	 * 高温の建物の周りを探索<br>
	 * 燃えてる建物を見つけやすくする
	 * @throws ActionCommandException
	 */
	protected void searchAroundHotBuilding()throws ActionCommandException{
		Building hotBuilding;
		hotBuilding=getHotBuilding();
		if(hotBuilding==null)return;
		
		int minTime=Integer.MAX_VALUE;
		int minDistance=Integer.MAX_VALUE;
		Building DoubtfulBuilding=null;
		
		//温度の高い建物から近い順に建物を6つ取ってくる
		Collection<StandardEntity> AroundHotBuildings = new LinkedList<StandardEntity>(getBuidingExtent(hotBuilding, 6));
		//周りに燃えてる建物があったならそこで終了
		if(AroundHotBuildings.removeAll(world.getBurningBuildings()))return;
		
		//TODO rangeの値は検討の余地あり
		//最も前に見た建物に入る
		for(StandardEntity se : AroundHotBuildings){
			if(se instanceof Building 
					&& !(se instanceof GasStation)
					&& !(se instanceof Refuge)
					&& !(se instanceof FireStation)
					&& !(se instanceof PoliceOffice)
					&& !(se instanceof AmbulanceCentre)) {
				//TODO costfunctionもしくは隣接するRoadなどの値から見ることが困難かを見よう
				int LastSeenTime = world.getTimestamp().getLastSeenTime(se.getID());
			
				if((!world.getTimestamp().haveSeen(se.getID()) || 20 < time - LastSeenTime)
						&& (LastSeenTime < minTime || (LastSeenTime==minTime && world.getDistance(hotBuilding, se)<minDistance) )
						&& router.isReachable((Area)se)){
					DoubtfulBuilding = (Building)se;
					minTime = world.getTimestamp().getLastSeenTime(se.getID());
					minDistance = world.getDistance(hotBuilding, se);
				}
			}
		}
		
		if(DoubtfulBuilding!=null){
			move(DoubtfulBuilding);
		}
		return ;
	}
	
	
	@Override
	public String toString() {
		return "Ri-one ambulance team agent";
	}
}
