package rione.agent.fb;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Hydrant;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.agent.PlatoonAgent;

import java.awt.geom.Ellipse2D;
import java.util.EnumSet;

public abstract class AbstractFireBrigadeAgent extends PlatoonAgent<FireBrigade> {

	protected void extinguish(EntityID target, int water) throws  ActionCommandException {
		sendExtinguish(time, target, water);
		throw new ActionCommandException(StandardMessageURN.AK_EXTINGUISH);
	}

	protected void extinguish(Building target, int water) throws  ActionCommandException {
		extinguish(target.getID(), water);
	}

	protected void extinguish(EntityID target) throws  ActionCommandException {
		extinguish(target, world.getConfig().maxPower);
	}

	protected void extinguish(Building target) throws  ActionCommandException {
		extinguish(target.getID());
	}

	/**
	 * Buildingが消火可能範囲にあるかどうかを返す<br>
	 * 他のBuildingにさえぎられ直接視認できないBuildingは、
	 * 消火後も気づかない可能性がある
	 * @param building 判断するBuilding
	 * @return 消火可能ならtrue
	 */
	protected boolean isExtinguishable(Pair<Integer,Integer> position , Building building) {
		final int r = world.getConfig().extinguishableDistance;
		java.awt.geom.Area range = new java.awt.geom.Area(new Ellipse2D.Double(
				position.first() - r, position.second() - r,
				r * 2, r * 2));
		if(building.getShape()!=null){
			range.intersect(new java.awt.geom.Area(building.getShape()));
		}
		return !range.isEmpty();
	}
	/**
	 * Buildingが消火可能範囲にあるかどうかを返す
	 * @param fb
	 * @param building
	 * @return
	 */
	protected boolean isExtinguishable(FireBrigade fb, Building building) {
		//Pair<Integer, Integer> position = fb.getLocation(world);
		Pair<Integer, Integer> position = new Pair<Integer,Integer>(fb.getX(),fb.getY());
		return isExtinguishable(position,building);
	}
	protected boolean isExtinguishable(Hydrant hydrant, Building building) {
		Pair<Integer, Integer> position = hydrant.getLocation(world);
		return isExtinguishable(position,building);
	}

	/**
	 * Buildingが消火可能範囲にあるかどうかを返す<br>
	 * 他のBuildingにさえぎられ直接視認できないBuildingは、
	 * 消火後も気づかない可能性がある
	 * @param building 判断するBuilding
	 * @return 消火可能ならtrue
	 */
	protected boolean isExtinguishable(Building building) {
		return isExtinguishable(me(), building);
	}

	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
	}
}
