package rione.agent.fb;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import com.sun.org.apache.bcel.internal.generic.GETSTATIC;

import rescuecore2.messages.Command;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.gui.ScreenTransform;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.agent.DebugAgent;
import rione.model.AdvancedWorldModel;
import rione.model.AgentConstants;
import rione.model.route.pov.CostFunction;
import rione.viewer.component.AdvancedViewComponent;
import rione.viewer.component.TransformedGraphics;
import rione.viewer.component.extension.ExtensionMap;
import rione.viewer.component.extension.TextExtension;

/**
 * デバッグ時は実行の構成の起動オプションからこちらを起動する
 * @author utisam
 *
 */
public class DebugFireBrigadeAgent extends FireBrigadeAgent implements DebugAgent {

	/** 最後に呼び出されたメソッドを表す文字列 */
	private String lastCommand = null;

	@Override
	protected void initialize() {
		super.initialize();
		//ビューアに登録
		AdvancedViewComponent.addTeamAgent(this);
	}

	@Override
	public AdvancedWorldModel getWorld() {
		return world;
	}
	
	//private DynamicConstant enableReachableExtension = DynamicConstant.create("'ReachableExtension(fb)', false");
	
	/*
	 * 自由に設定できるEntityIDと装飾の対応
	 * (非 Javadoc)
	 * @see rione.viewer.component.DebugAgent#customExtension()
	 */
	@Override
	public ExtensionMap customExtension() {
		ExtensionMap exMap = new ExtensionMap();
		world.getBurningBuildings().extensions(exMap);
		world.getBuriedHumans().extension(exMap);

		//router.reachableExtension(exMap);
		//world.getEnergyFlow().extension(exMap);
		
		for (StandardEntity se : world.getEntitiesOfType(AgentConstants.HUMANOIDS)) {
			Human human = (Human)se;
			//クリックしたAgentを表示
			if(se.getID() == me().getID()){
				exMap.put(se.getID(), new TextExtension(exMap.get(se.getID()),
						"ID: " + me().getID(), Color.YELLOW));
				if (human.isHPDefined() && human.isDamageDefined()){ 
					exMap.put(se.getID(), new TextExtension(exMap.get(se.getID()),
							"HP: " + human.getHP(), Color.YELLOW));
					if(human.getDamage() > 0)
					exMap.put(se.getID(), new TextExtension(exMap.get(se.getID()),
							"Damage: " + human.getDamage(), Color.YELLOW));
				}
				FireBrigade fb = (FireBrigade) se;
				if(fb.isWaterDefined()){
					exMap.put(se.getID(), new TextExtension(exMap.get(se.getID()),
							"Water: " + fb.getWater(), Color.YELLOW));
				}
			}
		}
		for(Building building : world.getBurningBuildings()){
			if(building.isTemperatureDefined())
				exMap.put(building.getID(), new TextExtension(exMap.get(building.getID()),
						"Temperature: " + building.getTemperature(), Color.YELLOW));
		}
		/*for(StandardEntity road : unlookupedRoads){
			exMap.put(road, new BorderExtension(exMap.get(road),Color.blue, 3.0f));
		}*/
		/*for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.GAS_STATION)) {
			for(StandardEntity seArea : world.getNeighbours(se, 10)){
				exMap.put(seArea, new BorderExtension(exMap.get(seArea),Color.blue, 3.0f));
			}
		}*/
		return exMap;
	}
	
	
	@Override
	public void customRender(Graphics2D g, ScreenTransform t, int width, int height) {
		g.setFont(new Font(g.getFont().getFontName(),Font.BOLD,20));
		TransformedGraphics gt = new TransformedGraphics(g, t);
		gt.setThickness(4);
		gt.setColor(Color.YELLOW);
		int maxTemp = 0;
		Building hotBuilding=null;
		//視界内の一番近い高温の建物を探す
		for(EntityID eid : getVisibleEntities()){
			StandardEntity se = world.getEntity(eid);
			if(se instanceof Building){
				Building bld = (Building) se;
				int temp = bld.getTemperature();
				if(bld.getFierynessEnum() == Fieryness.UNBURNT && 0<temp && maxTemp<temp){
					hotBuilding=(Building)se;
					maxTemp=temp;
				}else if(bld.getFierynessEnum() == Fieryness.BURNING){
					return;
				}
			}
		}
		if(hotBuilding==null)return;

		
		Collection<StandardEntity> AroundHotBuildings = new LinkedList<StandardEntity>(world.getObjectsInRange(hotBuilding, 5000));
		//周りに燃えてる建物があったならそこで終了
		if(AroundHotBuildings.removeAll(world.getBurningBuildings()))return;
		
		for(StandardEntity se : AroundHotBuildings){
			if(se instanceof Building){
				gt.draw(((Building) se).getShape());
			}
		}
		/*g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		TransformedGraphics gt = new TransformedGraphics(g, t);
		
		g.setFont(new Font(g.getFont().getFontName(),Font.BOLD,20));
		gt.setThickness(4);
		
		Building minDifficultyBuilding = null;
		double minDifficulty = Integer.MAX_VALUE;
		double minDifficultyAffect = 0.0;

		for (EntityID id : getVisibleEntities()) {
			StandardEntity se = world.getEntity(id);
			if (se instanceof Building && !(se instanceof GasStation)) {
				Building building = (Building) se;
				double area = (building.isTotalAreaDefined()) ? building.getTotalArea() : 1.0;
				double affected = world.getEnergyFlow().getIn(building);
				double difficulty = area * affected;
				if(!isExtinguishable(building))continue;
				if (world.getBurningBuildings().contains(building) && difficulty < minDifficulty) {
					minDifficultyBuilding = building;
					minDifficulty = difficulty;
					minDifficultyAffect = world.getEnergyFlow().getOut(building);
				}
				else if (Math.abs(minDifficulty - difficulty) < 500.0) {
					// difficultyが近ければaffectを比べる
					double affect = world.getEnergyFlow().getOut(building);
					if (minDifficultyAffect < affect) {
						minDifficultyBuilding = building;
						minDifficulty = difficulty;
						minDifficultyAffect = world.getEnergyFlow().getOut(building);
					}
				}
			}
		}
		
		double minDistance = Double.MAX_VALUE;
		Building minDistanceBuilding = null;
		Collection<StandardEntity> aroundEntity = new HashSet<StandardEntity>();
		if(minDifficultyBuilding != null){
			Collection<EntityID> r = new HashSet<EntityID>(minDifficultyBuilding.getNeighbours());
			
			for(EntityID e : r){
				if(world.getEntity(e) instanceof Road){
					Collection<Building> neightbourBuilding = new HashSet<Building>(getNeighbourBuilding((Road)world.getEntity(e), 5));
					if(!neightbourBuilding.isEmpty())aroundEntity.addAll(neightbourBuilding);
				}
			}
		}
		for(StandardEntity se : aroundEntity){
			gt.draw(((Building)se).getShape());
		}*/
	}
	
	private void energyflowDrawer(TransformedGraphics gt) {	
		gt.setColor(Color.WHITE);
		gt.setThickness(1.0f);
		final int rayCount = 16;
		
		for (StandardEntity se : world.getBurningBuildings()) {
			Building target = (Building) se;
			double r = Math.sqrt(target.getGroundArea()) * 1000.0 * 2.0;
			Collection<StandardEntity> range = world.getObjectsInRange(target, (int) r);
			double dAngle = Math.PI * 2 / rayCount;
			for (int i = 0; i < rayCount; ++i) {
				double angle = i * dAngle;
				double dx = Math.sin(angle) * r;
				double dy = Math.cos(angle) * r;
				Line2D ray = new Line2D(target.getX(), target.getY(), dx, dy);
				// 線分rayの中で最も近くに当たっているBuildingを求める
				Building nearestHit = null;
				double nearestHitDist = Double.MAX_VALUE;
				for (StandardEntity ase : range) {
					if (!(ase instanceof Building) || target.equals(ase))
						continue;
					final Building building = (Building) ase;
					for (Edge e : building.getEdges()) {
						if (e.isPassable()) continue;
						final Line2D l = e.getLine();
						final double d1 = ray.getIntersection(l);
						final double d2 = l.getIntersection(ray);
						if (d1 < nearestHitDist
								&& 0.0 <= d2 && d2 <= 1.0
								&& 0.0 < d1 && d1 <= 1.0) {
							nearestHit = building;
							nearestHitDist = d1;
						}
					}
				}
				if (nearestHit != null) {
					Point2D o = ray.getOrigin();
					Point2D pt = ray.getPoint(nearestHitDist);
					gt.drawLine((int) o.getX(), (int) o.getY(), (int) pt.getX(), (int) pt.getY());
				}
			}
		}
		
	}
	

	  //_/_/_/_/_/_/_/_/_/_/_/_/_/
	 //_/ 以下CommandCall用 /_/
	//_/_/_/_/_/_/_/_/_/_/_/_/
	
	@Override
	public String getCommandsCall() {
		if (lastCommand == null) {
			return "";
		}
		return lastCommand;
	}
	
	@Override
	protected void think(int time, ChangeSet changed, Collection<Command> heard) {
		final long thinkStart = System.currentTimeMillis();
		super.think(time, changed, heard);
		lastCommand += ": " + (System.currentTimeMillis() - thinkStart) + "[ms]";
	}
	
	@Override
	protected void act() throws ActionCommandException {
		lastCommand = "act";
		super.act();
	}

	// 以下自動生成
	 
	@Override
	protected void rest() throws ActionCommandException {
		lastCommand += "->rest";
		super.rest();
	}
	@Override
	protected void extinguish(EntityID target,int water) throws ActionCommandException {
		lastCommand += "->extinguish";
		super.extinguish(target,water);
	}
	@Override
	protected void moveFront(StandardEntity destination, CostFunction costFunc) throws ActionCommandException {
		lastCommand += "->moveFront";
		super.moveFront(destination, costFunc);
	}
	@Override
	protected void move(List<EntityID> path, int x, int y) throws ActionCommandException {
		lastCommand += "->move(path,x,y)";
		super.move(path, x, y);
	}
	@Override
	protected void moveFront(Collection<? extends StandardEntity> destinations) throws ActionCommandException {
		lastCommand += "->moveFront";
		super.moveFront(destinations);
	}
	@Override
	protected void moveFront(Collection<? extends StandardEntity> destinations, CostFunction costFunc) throws ActionCommandException {
		lastCommand += "->moveFront";
		super.moveFront(destinations, costFunc);
	}
	@Override
	protected void moveFront(List<EntityID> path) throws ActionCommandException {
		lastCommand += "->moveFront";
		super.moveFront(path);
	}
	@Override
	protected void moveFront(List<EntityID> path, int x, int y) throws ActionCommandException {
		lastCommand += "->moveFront";
		super.moveFront(path, x, y);
	}
	@Override
	protected void move(StandardEntity destination, CostFunction costFunc) throws ActionCommandException {
		lastCommand += "->move(des,cost)";
		super.move(destination, costFunc);
	}
	@Override
	protected void move(StandardEntity destination) throws ActionCommandException {
		lastCommand += "->move(des)";
		super.move(destination);
	}
	@Override
	protected void moveReachable(Collection<? extends StandardEntity> destinations, boolean front) throws ActionCommandException {
		lastCommand += "->moveReachable";
		super.moveReachable(destinations, front);
	}
	@Override
	protected void moveReachable(Collection<? extends StandardEntity> destinations, CostFunction costFunc, boolean front) throws ActionCommandException {
		lastCommand += "->moveReachable";
		super.moveReachable(destinations, costFunc, front);
	}
	@Override
	protected void moveReachable(List<EntityID> path, boolean front) throws ActionCommandException {
		lastCommand += "->moveReachable";
		super.moveReachable(path, front);
	}
	@Override
	protected void moveReachable(List<EntityID> path, int x, int y, boolean front) throws ActionCommandException {
		lastCommand += "->moveReachable";
		super.moveReachable(path, x, y, front);
	}
	@Override
	protected void goOut(Building bld) throws ActionCommandException {
		lastCommand += "->goOut";
		super.goOut(bld);
	}
	@Override
	protected void lookupSearchBuildings() throws ActionCommandException {
		lastCommand += "->lookupSearchBuildings";
		super.lookupSearchBuildings();
	}
	@Override
	protected void enterSearchBuildings() throws ActionCommandException {
		lastCommand += "->enterSearchBuildings";
		super.enterSearchBuildings();
	}
	@Override
	protected void randomWalk() throws ActionCommandException {
		lastCommand += "->randomWalk";
		super.randomWalk();
	}
	@Override
	protected void moveToRefuge() throws ActionCommandException {
		lastCommand += "->moveToRefuge";
		super.moveToRefuge();
	}
	@Override
	protected void aggregatorStay() throws ActionCommandException {
		lastCommand += "->aggregatorStay";
		super.aggregatorStay();
	}
	@Override
	protected void messengerLoop(int interval) throws ActionCommandException {
		lastCommand += "->messengerLoop";
		super.messengerLoop(interval);
	}
	@Override
	protected void giveUpPosition() throws ActionCommandException {
		lastCommand += "->giveUpPosition";
		super.giveUpPosition();
	}
	@Override
	protected void checkAreas() throws ActionCommandException {
		lastCommand += "->checkAreas";
		super.checkAreas();
	}
	@Override
	protected void moveToPartedArea() throws ActionCommandException {
		lastCommand += "->moveToPartedArea";
		super.moveToPartedArea();
	}
	@Override
	protected void moveToFires() throws ActionCommandException {
		lastCommand += "->moveToFires";
		super.moveToFires();
	}
	@Override
	protected void visibleExtingish() throws ActionCommandException {
		lastCommand += "->visibleExtingish";
		super.visibleExtingish();
	}
	@Override
	protected void radioExtingish() throws ActionCommandException {
		lastCommand += "->radioExtingish";
		super.radioExtingish();
	}
	@Override
	protected void supplyWater() throws ActionCommandException {
		lastCommand += "->supplyWater";
		super.supplyWater();
	}
	@Override
	protected void complexSupplyWater() throws ActionCommandException {
		lastCommand += "->complexSupplyWater";
		super.complexSupplyWater();
	}
	@Override
	protected void lookupSearchRoads() throws ActionCommandException {
		lastCommand += "->lookupSearchRoads";
		super.lookupSearchRoads();
	}
	@Override
	protected void cantMove() throws ActionCommandException {
		lastCommand += "->cantMove";
		super.cantMove();
	}
	@Override
	protected void errorExtinguish() throws ActionCommandException {
		lastCommand += "->errorExtinguish";
		super.errorExtinguish();
	}
	/*@Override
	protected void careSelf() throws ActionCommandException {
		lastCommand += "->careSelf";
		super.careSelf();
	}*/
}
