package rione.agent.fb;

import java.awt.Polygon; 
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Hydrant;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rione.model.AgentConstants;
import rione.model.areapartition.PartedArea;
import rione.model.areapartition.PartedArea.AreaStatus;

public class FireBrigadeAgent extends AbstractFireBrigadeAgent {

	ArrayList<FireBrigade> fbsMaster = new ArrayList<FireBrigade>();

	@Override
	protected void initialize(){
		super.initialize();
		System.out.println(toString() + " was connected. [id=" + getID() + ", uniform=" + getUniform() + "]");
		for(StandardEntity se : world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)){
			FireBrigade fb = (FireBrigade) se;
			fbsMaster.add(fb);
		}
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();
	}

	@Override
	protected void act() throws ActionCommandException {
		
		cantMove();
		careSelf();
		supplyWater();
		giveUpPosition();
		extinguishPosition();
		visibleExtingish();
		if (!isNoRadio()) {
			radioExtingish();
		}
		if (isNoRadio() && world.getBurningBuildings().isEmpty() && time > 25) {
			checkAggregator();
		}
		searchAroundHotBuilding();
		moveToFires();
		visibleInfernoExtinguish();
		moveToInfernoFires();
		lookupSearchRoads();
		randomWalk();
	}

	private static final EnumSet<Fieryness> giveUpFieryness = EnumSet.of(
			Fieryness.BURNT_OUT,
			Fieryness.SEVERE_DAMAGE,
			Fieryness.INFERNO
			);
	private static final EnumSet<Fieryness> dangerFieryness = EnumSet.of(
			Fieryness.BURNT_OUT, 
			Fieryness.BURNING
			);

	/**
	 * 視界内がすべてgiveUpFierynessだった場合にはあきらめる
	 * @throws ActionCommandException
	 */
	protected void giveUpPosition() throws ActionCommandException {
		if (isAllGiveUp()) return;
		for (EntityID id : getVisibleEntities()) {
			StandardEntity se = world.getEntity(id);
			if (se.getStandardURN() == StandardEntityURN.BUILDING) {
				Building building = (Building) se;
				if (!giveUpFieryness.contains(building.getFierynessEnum())) {
					// あきらめる必要のないBuildingが視界内に含まれている
					return;
				}
			}
		}
		moveToFires();
	}
	private boolean isAllGiveUp() {
		for (Building building : world.getBurningBuildings()) {
			if (!giveUpFieryness.contains(building.getFierynessEnum())) {
				return true;
			}
		}
		return false;
	}

	/**
	 * ガソリンスタンドの発火を止められないなら消火する
	 * @throws ActionCommandException
	 */
	protected void giveUpGas() throws ActionCommandException {
		boolean judge = false;
		GasStation target = null;
		int bld=0;
		int agent=0;
		for(EntityID id : getVisibleEntities()){
			StandardEntity se = world.getEntity(id);
			if (se instanceof GasStation && ((GasStation) se).isFierynessDefined()
					&& ((GasStation)se).getFierynessEnum().equals(Fieryness.UNBURNT)) {
				judge=true;
				target = (GasStation)se;
			}else if(se instanceof Building && world.getBurningBuildings().contains((Building)se)){
				bld++;
			}else if(se instanceof AmbulanceTeam || se instanceof FireBrigade ||
					se instanceof PoliceForce){
				agent++;
			}
		}
		if(judge==true && bld > 2 && agent < 4){
			extinguish(target);
		}
	}

	private PartedArea decideMyArea() {
		HashSet<PartedArea> burningPA = new HashSet<PartedArea>();
		for (PartedArea pa : world.getAreaPartitions()) {
			if (pa.getStatus().equals(AreaStatus.burning)) {
				burningPA.add(pa);
			}
		}
		if (burningPA.isEmpty()) {
			return null;
		}

		ArrayList<FireBrigade> fbs = new ArrayList<FireBrigade>(fbsMaster);
		for (Iterator<FireBrigade> it = fbs.iterator(); it.hasNext();) {
			if (!it.next().isPositionDefined()) {
				it.remove();
			}
		}

		for (final PartedArea pa : burningPA) {
			if (fbs.isEmpty()) {
				return null;
			}
			Collections.sort(fbs, new Comparator<FireBrigade>() {
				@Override
				public int compare(FireBrigade o1, FireBrigade o2) {
					double dist1 = Point2D.distance(
							((Area) o1.getPosition(world)).getX(),
							((Area) o1.getPosition(world)).getY(),
							pa.getPoints().getX(),
							pa.getPoints().getY());
					double dist2 = Point2D.distance(
							((Area) o2.getPosition(world)).getX(),
							((Area) o2.getPosition(world)).getY(),
							pa.getPoints().getX(), pa.getPoints().getY());
					if (dist1 < dist2) return -1;
					if (dist1 > dist2) return 1;
					return 0;
				}
			});
			// TODO PartedAreaに対する人数割り当て
			int paNum = fbs.size() / burningPA.size() > 1 ? fbs.size() / burningPA.size() : 1;
			Iterator<FireBrigade> it = fbs.iterator();
			for (int i = 0; it.hasNext() && i < paNum; i++) {
				if (it.next().getID().equals(me().getID())) {
					return pa;
				}
				else {
					it.remove();
				}
			}
		}
		return null;
	}

	protected void checkAggregator() throws rione.agent.Agent.ActionCommandException {
		int lastTime = -1;
		EntityID lastSeenID = null;
		for (EntityID id : getAggregators()) {
			int t = world.getTimestamp().getLastSeenTime(id);
			if (lastTime < t) {
				lastTime = t;
				lastSeenID = id;
			}
		}
		if (lastTime == -1) {
			Set<Area> aggregatorsArea = new HashSet<Area>(getAggregators().size());
			for (EntityID id : getAggregators()) {
				aggregatorsArea.add((Area) world.getEntity(getAggregatorPosition(id)));
			}
			move(aggregatorsArea);
		}
		if (lastTime - time > 10) {
			move(lastSeenID);
		}
	}

	HashSet<Building> needCheckBuildings = new HashSet<Building>();
	protected void checkAreas() throws ActionCommandException {
		for(PartedArea pa: world.getAreaPartitions()){
			if(pa.isStateChanged() && pa.getStatus().equals(AreaStatus.extinguished)){
				for(EntityID id : pa.getContentBuildingsID()){
					needCheckBuildings.add((Building) world.getEntity(id));
				}
			}
		}
		if(!needCheckBuildings.isEmpty()){
			for(Iterator<Building> it = needCheckBuildings.iterator() ; it.hasNext();){
				Building bld = it.next();
				if(world.getVisibleEntitiesID().contains(bld.getID())
						|| (bld.isFierynessDefined() && bld.getFierynessEnum().equals(Fieryness.BURNT_OUT))
						|| !router.isSureReachable(bld)){
					it.remove();
				}
			}
			for(StandardEntity se : world.getEntitiesOfType(AgentConstants.PLATOONS)){
				Human human = (Human) se;
				if(human.isPositionDefined() && world.getVisibleEntitiesID().contains(human.getPosition())){
					needCheckBuildings.remove(human.getPosition(world));
				}
			}
			if(!needCheckBuildings.isEmpty()){
				moveFront(needCheckBuildings);
			}
		}

	}

	protected void moveToPartedArea() throws ActionCommandException {
		final PartedArea myPartition = decideMyArea();
		if (myPartition != null) {
			if (!myPartition.containAtLeastOne(world.getVisibleEntitiesID())) {
				HashSet<StandardEntity> noBurnings = new HashSet<StandardEntity>();
				HashSet<Building> burnings = new HashSet<Building>();
				for (EntityID id : myPartition.getContentBuildingsID()) {
					Building building = (Building) world.getEntity(id);
					if (building.isOnFire()) {
						burnings.add(building);
					}
					noBurnings.add(building);
				}
				if (burnings.isEmpty()) {
					move(noBurnings);
				}
				else {
					moveReachable(burnings, true);
				}
			}
		}

	}

	/**
	 * 燃えている建物に移動する
	 * @throws ActionCommandException
	 */
	protected void moveToFires() throws ActionCommandException {
		if (world.getBurningBuildings().isEmpty()) return;

		Building maxValueBuilding = null;
		double maxValue = Integer.MIN_VALUE;

		for (Building building : world.getBurningBuildings()) {
			if(building.getFierynessEnum() == StandardEntityConstants.Fieryness.INFERNO) continue;
			// TODO 実験の結果与える影響が大きい方がいいらしい．
			// 感覚と違うので他のマップで要検証
			final double affect = world.getEnergyFlow().getOut(building);
			final double distance = Math.max(1.0, world.getDistance(building, me()));
			//TODO 重み付きで考えたほうが良いかもしれない
			final double value = affect / Math.sqrt((double)distance);
			if (maxValue < value) {
				maxValueBuilding = building;
				maxValue = value;
			}
			//TODO minValueより多少大きくても，周囲に与える影響(Out)が大きければこちらにする
		}
		if (maxValueBuilding != null) {
			moveFront(maxValueBuilding, router.getFbCostFunction(maxValueBuilding));
		}
	}

	protected void moveToInfernoFires() throws rione.agent.Agent.ActionCommandException {
		if (world.getBurningBuildings().isEmpty()) return;

		Building maxValueBuilding = null;
		double maxValue = Integer.MIN_VALUE;

		for (Building building : world.getBurningBuildings()) {
			// TODO 実験の結果与える影響が大きい方がいいらしい．
			// 感覚と違うので他のマップで要検証
			final double affect = world.getEnergyFlow().getOut(building) + world.getEnergyFlow().getIn(building);
			final double distance = Math.max(1.0, world.getDistance(building, me()));
			//TODO 重み付きで考えたほうが良いかもしれない
			final double value = affect / Math.sqrt((double)distance);
			if (maxValue < value) {
				maxValueBuilding = building;
				maxValue = value;
			}
			//TODO minValueより多少大きくても，周囲に与える影響(Out)が大きければこちらにする
		}
		if (maxValueBuilding != null) {
			moveFront(maxValueBuilding, router.getFbCostFunction(maxValueBuilding));
		}
	}

	/**
	 * 視界内の建物を消火する
	 * @throws ActionCommandException
	 */
	protected void visibleExtingish() throws ActionCommandException {
		Building minDifficultyBuilding = null;
		double minDifficulty = Integer.MAX_VALUE;
		double minDifficultyAffect = 0.0;

		for (EntityID id : getVisibleEntities()) {
			StandardEntity se = world.getEntity(id);
			if (se instanceof Building && !(se instanceof GasStation)) {
				Building building = (Building) se;
				if(building.getFierynessEnum() == StandardEntityConstants.Fieryness.INFERNO) continue;
				double area = (building.isTotalAreaDefined()) ? building.getTotalArea() : 1.0;
				double affected = world.getEnergyFlow().getOut(building);
				double difficulty = area * affected;
				if(!isExtinguishable(building))continue;
				if (world.getBurningBuildings().contains(building) && difficulty < minDifficulty) {
					minDifficultyBuilding = building;
					minDifficulty = difficulty;
					minDifficultyAffect = world.getEnergyFlow().getOut(building);
				}
				else if (Math.abs(minDifficulty - difficulty) < 500.0) {
					// difficultyが近ければaffectを比べる
					double affect = world.getEnergyFlow().getOut(building);
					if (minDifficultyAffect < affect) {
						minDifficultyBuilding = building;
						minDifficulty = difficulty;
						minDifficultyAffect = world.getEnergyFlow().getOut(building);
					}
				}
			}
		}

		if (minDifficultyBuilding != null) {
			extinguish(minDifficultyBuilding);
		}
	}

	protected void visibleInfernoExtinguish() throws rione.agent.Agent.ActionCommandException{
		Building minDifficultyBuilding = null;
		double minDifficulty = Integer.MAX_VALUE;
		double minDifficultyAffect = 0.0;

		for (EntityID id : getVisibleEntities()) {
			StandardEntity se = world.getEntity(id);
			if (se instanceof Building && !(se instanceof GasStation)) {
				Building building = (Building) se;
				double area = (building.isTotalAreaDefined()) ? building.getTotalArea() : 1.0;
				double affected = world.getEnergyFlow().getIn(building) + world.getEnergyFlow().getOut(building);
				double difficulty = area * affected;
				if(!isExtinguishable(building))continue;
				if (world.getBurningBuildings().contains(building) && difficulty < minDifficulty) {
					minDifficultyBuilding = building;
					minDifficulty = difficulty;
					minDifficultyAffect = world.getEnergyFlow().getOut(building);
				}
				else if (Math.abs(minDifficulty - difficulty) < 500.0) {
					// difficultyが近ければaffectを比べる
					double affect = world.getEnergyFlow().getOut(building);
					if (minDifficultyAffect < affect) {
						minDifficultyBuilding = building;
						minDifficulty = difficulty;
						minDifficultyAffect = world.getEnergyFlow().getOut(building);
					}
				}
			}
		}

		if (minDifficultyBuilding != null) {
			extinguish(minDifficultyBuilding);
		}	
	}

	/**
	 * 建物を消火可能な範囲の近くの消火栓をとってくる
	 * @param 消火中の建物
	 * @return 消火栓
	 */
	protected Hydrant getNeighborHydrant(Building bld){
		Hydrant nearestHydrant=null;
		int nearestDistance=Integer.MAX_VALUE;
		Collection<StandardEntity> hydrants = world.getEntitiesOfType(StandardEntityURN.HYDRANT);
		for(StandardEntity hyd : hydrants){
			if(isExtinguishable(hyd.getLocation(world), bld)){
				if(nearestHydrant==null || nearestDistance > world.getDistance(me(), hyd)){
					nearestHydrant=(Hydrant)hyd;
				}
			}
		}
		if(nearestHydrant!=null){
			return nearestHydrant;
		}
		return null;
	}

	/**
	 * 視界外の建物も含めて消火する．
	 * @throws rione.agent.Agent.ActionCommandException
	 */
	protected void radioExtingish() throws ActionCommandException {
		
		final int t = (world.getLocateInBlockade().isLocateInBlockade(me().getLocation(world)) != null)
				? 8 : 3;
		//		int minFieryness = StandardEntityConstants.Fieryness.values().length;
		//		Building minFierynessBuilding = null;
		//		for(Building building : world.getBurningBuildings()){
		//			// 消えても放水し続ける可能性があるのでtimestampから確認する
		//			if (isExtinguishable(building)
		//					&& world.getTime() - world.getTimestamp().getLastChangedTime(building.getID()) < t){
		//				if (building.isFierynessDefined() && minFieryness > building.getFieryness()) {
		//					minFierynessBuilding = building;
		//					minFieryness = building.getFieryness();
		//				}
		//			}
		//		}
		//		if (minFierynessBuilding != null) {
		//			extinguish(minFierynessBuilding);
		//		}
		Building minDifficultyBuilding = null;
		double minDifficulty = Integer.MAX_VALUE;
		double minDifficultyAffect = 0.0;

		for (Building building : world.getBurningBuildings()) {
			if (!isExtinguishable(building)
					|| world.getTime() - world.getTimestamp().getLastChangedTime(building.getID()) >= t) {
				continue;
			}
			if (building.isFierynessDefined() && building.getFierynessEnum() == Fieryness.INFERNO) {
				continue;
			}
			double area = (building.isTotalAreaDefined()) ? building.getTotalArea() : 1.0;
			double affected = world.getEnergyFlow().getIn(building);
			double difficulty = area * affected;
			if (world.getBurningBuildings().contains(building) && difficulty < minDifficulty) {
				minDifficultyBuilding = building;
				minDifficulty = difficulty;
				minDifficultyAffect = world.getEnergyFlow().getOut(building);
			}
			else if (Math.abs(minDifficulty - difficulty) < 500.0) {
				// difficultyが近ければaffectを比べる
				double affect = world.getEnergyFlow().getOut(building);
				if (minDifficultyAffect < affect) {
					minDifficultyBuilding = building;
					minDifficulty = difficulty;
					minDifficultyAffect = world.getEnergyFlow().getOut(building);
				}
			}
		}
		if (minDifficultyBuilding != null){
			extinguish(minDifficultyBuilding);
		}
	}

	protected void supplyWater() throws ActionCommandException {
		final int maxPower = world.getConfig().maxPower;
		final int remain = world.getConfig().timestep - time;
		final int capacity = world.getConfig().maxTankCapacity > maxPower ? world.getConfig().maxTankCapacity:maxPower;
		final int refillRate = world.getConfig().tankRefillRate;
		final int currentWater = me().getWater();
		// 最大放水を繰り返しても残るなら給水しない
		// 満タンなら給水しない
		if (currentWater - remain * maxPower > 2 || me().getWater() >= capacity){
			return;
		}
		// 継続判定
		if(location() instanceof Refuge) {
			// refillRateを有効活用できなさそうなら出る
			if (currentWater + refillRate > capacity){
				return;
			}
			rest();
		}
		// ルール変更によってキャパシティが小さすぎる(capacity < maxPower)場合
		// も考えられるので，なるべく対応する
		// 最大で放水できなければ給水
		if (currentWater < Math.min(capacity, maxPower)) {
			moveToRefuge();
		}
		complexSupplyWater();
	}

	protected void complexSupplyWater() throws ActionCommandException {
		int fbnum = world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE).size();
		// 同時に給水する最大人数
		int supplyFB = fbnum / 3;

		if (world.getBurningBuildings().isEmpty()
				&& me().equals(getMinTank())
				&& refugeFBcount() < supplyFB) {
			moveToRefuge();
		}
	}

	private FireBrigade getMinTank() {
		FireBrigade minFB = me();
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)) {
			FireBrigade fb = (FireBrigade) se;
			if (fb.isWaterDefined() && fb.getWater() < minFB.getWater()) {
				minFB = fb;
			}
		}
		return minFB;
	}

	/**
	 * Refuge内にいるFBの数を返す
	 * @return
	 */
	private int refugeFBcount(){
		int count = 0;
		for(EntityID id : getVisibleEntities()){
			Entity e = world.getEntity(id);
			if(e instanceof FireBrigade){
				FireBrigade fb = (FireBrigade)e;
				if(world.getEntity(fb.getPosition()) instanceof Refuge){
					count++;
				}
			}
		}
		return count;
	}

	protected void careSelf() throws ActionCommandException {
		//TODO Refugeまでの距離を考慮する
		if (me().getHP() - me().getDamage() * (timestep - time) < 0) {
			moveToRefuge();
		}
	}

	protected void cantMove() throws ActionCommandException {
		// moveしようとしているにも関わらず移動できていないとき
		// 建物から出られないときの対策
		if(getCommandHistory(time-1) != null && getCommandHistory(time-2) != null){
			if (getCommandHistory(time - 1).equals(StandardMessageURN.AK_MOVE)
					&& getCommandHistory(time - 2).equals(StandardMessageURN.AK_MOVE)) {
				if (locationHistory.get(time - 1).equals(locationHistory.get(time - 2))) {
					if(time % 4 == 0 && location() instanceof Building) {
						goOut((Building) location());
					}
					errorExtinguish();
					if(time % 3 == 0 && location() instanceof Road){
						randomWalk();
					}
				}
			}
		}
	}

	/**
	 * 手近な火災を消火
	 * @throws rione.agent.Agent.ActionCommandException
	 */
	protected void errorExtinguish() throws ActionCommandException {
		final int extinguishableDistance = world.getConfig().extinguishableDistance;
		for (StandardEntity se : world.getObjectsInRange(me(), extinguishableDistance)) {
			if (se instanceof Building && !(se instanceof GasStation)) {
				Building building = (Building) se;
				if (building.isOnFire() && me().isWaterDefined()) {
					if(isExtinguishable(building) && me().getWater() > world.getConfig().maxPower){
						extinguish(building);
					}
				}
			}
		}
	}

	/**
	 * 高温の建物の周りを探索<br>
	 * 燃えてる建物を見つけやすくする
	 * @throws ActionCommandException
	 */
	protected void searchAroundHotBuilding()throws ActionCommandException{
		Building hotBuilding;
		hotBuilding=getHotBuilding();
		if(hotBuilding==null)return;

		//温度の高い建物から近い順に建物を6つ取ってくる
		Collection<StandardEntity> AroundHotBuildings = new LinkedList<StandardEntity>(getBuidingExtent(hotBuilding, 6));
		//周りに燃えてる建物があったならそこで終了
		if(AroundHotBuildings.removeAll(world.getBurningBuildings()))return;

		int minTime=Integer.MAX_VALUE;
		int minDistance=Integer.MAX_VALUE;
		Building DoubtfulBuilding=null;



		//最も前に見た建物に入る
		for(StandardEntity se : AroundHotBuildings){
			//TODO costfunctionもしくは隣接するRoadなどの値から見ることが困難かで決めることができる可能性がある
			if(se instanceof Building 
					&& !(se instanceof GasStation)
					&& !(se instanceof Refuge)
					&& !(se instanceof FireStation)
					&& !(se instanceof PoliceOffice)
					&& !(se instanceof AmbulanceCentre)) {
				int LastSeenTime = world.getTimestamp().getLastSeenTime(se.getID());
				//最後に見たサイクルが同じの場合は距離で考える
				if((!world.getTimestamp().haveSeen(se.getID()) || 20 < time - LastSeenTime)
						&& (LastSeenTime < minTime || (LastSeenTime==minTime && world.getDistance(hotBuilding, se)<minDistance) )){
					if(router.isReachable((Area)se)){
						DoubtfulBuilding = (Building)se;
						minDistance = world.getDistance(hotBuilding, se);
						minTime = LastSeenTime;
					}
				}
			}
		}

		if(DoubtfulBuilding!=null){
			move(DoubtfulBuilding);
		}
	}
	/**
	 * 消火時の最適なポジションを決める
	 * A : 燃えている建物の集合
	 * B : 温度を保持しているがOnFireしていない建物の集合
	 * Bの建物の座標を頂点にして作った図形とAの建物の座標を頂点にして作った図形をつくる。
	 * Bの図形とAの図形の重複部分を除いたBの図形の中に存在するRoadを取得する。
	 * 取得されたRoadの中で１番近いRoadを消火時のポジションとする
	 * @throws rione.agent.Agent.ActionCommandException
	 * @author nanana
	 */
	private void extinguishPosition() throws rione.agent.Agent.ActionCommandException{
		List<Building> A = new ArrayList<Building>();	//TODO　名前変更		用途：放水範囲内の燃えているBuildingのList
		List<Building> B = new ArrayList<Building>();	//TODO	名前変更		用途：放水範囲内の燃えていない且つ、温度が0℃より高いBuildingのList
		List<Point2D>roads=new ArrayList<Point2D>();
		if(me().getWater()==0)return; //水が無いときはRefugeに向かうので処理はしない
		StandardEntity standardEntity = world.getEntity(me().getPosition()); 
		if(world.getEntitiesOfType(StandardEntityURN.REFUGE).contains(standardEntity))return; //Refugeにいるときは処理をしない。
		//視界内の集合を元に消火時のポジションを考える
		for(EntityID id:world.getVisibleEntitiesID()){
			StandardEntity se =world.getEntity(id);
			if(se instanceof Road){
				Road road=(Road)se;
				Point2D point2d = new Point2D.Double(road.getX(), road.getY());
				roads.add(point2d);
				continue;
			}
			
			if(se instanceof Building){
				Building building = (Building)se;
				if(!building.isFierynessDefined()||!building.isTemperatureDefined())continue;
				if(building.getTemperature()>0&&building.getFierynessEnum().equals(Fieryness.UNBURNT)){
					B.add(building);
				}else if(building.getFierynessEnum().equals(Fieryness.INFERNO)||building.getFierynessEnum().equals(Fieryness.BURNING)){
					A.add(building);
				}else if(building.getFierynessEnum().equals(Fieryness.WATER_DAMAGE)){
					B.add(building);
				}
			}
		}
		if(A.isEmpty()||B.isEmpty()){
			return;
		}
		int[] Ax = new int[A.size()]; //Aの集合内建物のx座標の配列
		int[] Ay = new int[A.size()]; //Aの集合内建物のy座標の配列
		int[] Bx = new int[B.size()]; //Bの集合内建物のx座標の配列
		int[] By = new int[B.size()]; //Bの集合内建物のy座標の配列
		int i=0;
		for(Building building:A){
			Ax[i]=building.getX();
			Ay[i]=building.getY();
			i++;
		}
		i=0;
		for(Building building:B){
			Bx[i]=building.getX();
			By[i]=building.getY();
			i++;
		}
		Polygon polygon1 = new Polygon(Ax, Ay, Ax.length); //Aの集合内の建物の座標を頂点とした図形
		Polygon polygon2 = new Polygon(Bx, By, Bx.length); //Bの集合内の建物の座標を頂点とした図形
		Rectangle range1 = polygon1.getBounds(); //このあとの処理を行うためにRectangle型に入れなおす 
		Rectangle range2 = polygon2.getBounds(); //このあとの処理を行うためにRectangle型に入れなおす
		range1.intersection(range2); //AとBの図形の重複部分をrange1に代入する処理
		List<Road>roads2=new ArrayList<Road>();
		roads2.clear();
		for(Point2D point2d:roads){ 
			if(range2.contains(point2d)&&!range1.contains(point2d)){ //Bの図形内に存在する且つAとBの重複部分には存在しない
				roads2.add((Road)world.getPositionFromCoordinates((int)point2d.getX(),(int) point2d.getY()));
			}
		}
		double distance= Integer.MAX_VALUE;
		Road road2=null;
		//選択されたRoadの中で自分との距離が１番近いRoadを選ぶ。
		for(Road road:roads2){
			if(distance>world.getDistance(me().getID(), road.getID())){
				distance=world.getDistance(me().getID(), road.getID());
				road2=road;
			}
		}
		if(road2==null)return;
		//選択されたRoadと自分の位置が同じ時は処理を終了
		if(me().getPosition().equals(road2.getID()))return;
		move(road2.getID());
	}
	@Override
	public String toString() {
		return "Ri-one fire brigade agent";
	}

}
