package rione.agent.pf;

import java.util.*;

import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.GasStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Hydrant;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.StandardMessageURN;
import rescuecore2.worldmodel.EntityID;
import rione.model.AgentConstants;
import rione.model.areapartition.PartedArea;
import rione.util.DistanceSorter;

public class PoliceForceAgent extends AbstractPoliceForceAgent {

	@Override
	protected void initialize() {
		super.initialize();

		// 接続表示
		System.out.println(toString() + " was connected. [id=" + getID() + ", uniform=" + getUniform() + "]");
	}

	@Override
	protected void prepareForAct() {
		super.prepareForAct();
		world.getCriticalArea().update(router);
	}

	@Override
	protected void act() throws ActionCommandException {
		cantMove();
		careSelf();
		meLocateInBlockade();
		cleanBuriedEntrance();
		clearNearLocateInBlockade();
		helpAgents();
		searchAroundHotBuilding();
		clearNearRefuge();
		clearRefuge();
		clearNearGasStation();	
		if (isAggregator()) {
			aggregatorStay();
		}
		else {
			if (isMessenger()) {
				messengerLoop(25);
			}
			prioritizeClear();
			if(getUniform()%2==0){
				clearAroundBurningBuilding();
			}
			clearBuriedHuman();
			if (time > world.getConfig().timestep / 2) {
				// 後半なら全て消して進む
				clearAround();
			}
			lookupSearchRoads();
			// TODO BurningBuildingsとFB間が!isSureReachableならその間のPathを開ける
			// TODO BuriedHumanとAT間が!isSureReachableならその間のPathを開ける
			if (time < 30) {
				clearCriticalArea();
			}
			searchFires();
			eraseUnreachable();
			randomWalk();
		}
	}

	/**
	 * 自分が埋まっているなら無理やり掘り起こす
	 * @throws ActionCommandException
	 */
	protected void meLocateInBlockade() throws ActionCommandException{
		if(time == 4){
			for (Human hm : world.getLocateInBlockade().getLocateInBlocades()) {
				if (hm.isPositionDefined()) {
					if(me().getID()==hm.getID()){
						clear(me().getX()-5000,me().getY()-5000);
					}
				}
			}
		}
		if(time == 5){
			for (Human hm : world.getLocateInBlockade().getLocateInBlocades()) {
				if (hm.isPositionDefined()) {
					if(me().getID()==hm.getID()){
						clear(me().getX()+5000,me().getY()+5000);
					}
				}
			}
		}
	}

	/**
	 * prioritize周辺の瓦礫を処理する
	 * @throws rione.agent.Agent.ActionCommandException
	 */
	protected void prioritizeClear() throws ActionCommandException{
		prioritize();
		if(PrioritizeBuilding.isEmpty() || world.getBuriedHumans() == null)return;
		int canadidate = getUniform() / 2;
		Collection<Building> buildings = PrioritizeBuilding;
		if (canadidate >= buildings.size())return;

		int min = Integer.MAX_VALUE;
		Building temp = null;
		Set<Building> removeBuildings = new HashSet<Building>();
		StandardEntity pos = world.getEntity(me().getPosition());
		//自分とPrioritizeATの集合の中で最も距離が近い建物を調べる
		for (Building building : PrioritizeBuilding) {
			int distance = world.getDistance(building, me());
			if(!router.isSureReachable(building) && distance < min && distance < getDistAverage() * 4){
				min = distance;
				temp = building;
			}
			if(pos.getID().equals(building.getID()) || router.isSureReachable(building)
					|| isVisible(building)){
				removeBuildings.add(building);
			}
		}
		if(!removeBuildings.isEmpty()){
			PrioritizeBuilding.removeAll(removeBuildings);
		}
		if(temp == null)return;
		move(temp, router.getPfCostFunction());
	}
	/**
	 * Buildingの前の道を見まわり
	 * 
	 * @throws ActionCommandException
	 */
	protected void lookupSearchRoads() throws ActionCommandException {
		int min = Integer.MAX_VALUE;
		StandardEntity temp = null;
		for(StandardEntity road : unlookupedRoadsInFrontOfTheBuilding){
			StandardEntity pos = world.getEntity(me().getPosition());
			int distance = world.getDistance(pos,road);
			Area area = (Area)road;
			if(!router.isSureReachable(area) && distance < min){
				min = distance;
				temp = road;
			}
		}
		if(temp != null){
			move(temp,router.getPfCostFunction());
		}
	}

	/**
	 * ガソリンスタンドに向かう
	 * @throws ActionCommandException
	 */
	protected void clearGasStation() throws ActionCommandException {
		int canadidate = getUniform() / 3;
		Collection<StandardEntity> gasstations = world.getEntitiesOfType(StandardEntityURN.GAS_STATION);
		if (canadidate < gasstations.size()) {
			EntityID id = world.getUniform().toID(StandardEntityURN.GAS_STATION, canadidate);
			GasStation assignedGas = (GasStation) world.getEntity(id);
			if (!router.isSureReachable(assignedGas)) {
				move(assignedGas, router.getPfCostFunction());
			}
		}
	}

	/**
	 * 近くのGasStationがReachableでなく，そのGasStationに一番近いなら，
	 * 割り当てられて無くても開ける．
	 * @throws ActionCommandException 
	 */
	protected void clearNearGasStation() throws ActionCommandException {
		Collection<StandardEntity> gass = world.getEntitiesOfType(StandardEntityURN.GAS_STATION);
		if (gass.isEmpty()) return;
		GasStation gas = (GasStation) Collections.min(gass, new DistanceSorter(me(), world));
		if (router.isSureReachable(gas)) return;

		PoliceForce nearest = me();
		int nearestDist = world.getDistance(gas, nearest);
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
			if (se.equals(nearest)) continue;
			PoliceForce pf = (PoliceForce) se;
			if (pf.isXDefined() && pf.isYDefined()) {
				int dist = world.getDistance(gas, pf);
				if (dist < nearestDist) {
					nearestDist = dist;
					nearest = pf;
				}
			}
		}
		if (me().equals(nearest)) {
			move(gas, router.getPfCostFunction());
		}
	}

	/**
	 * 消火栓に向かう
	 * @throws ActionCommandException
	 */
	protected void clearHydrant() throws ActionCommandException {
		int canadidate = getUniform() / 2;
		Collection<StandardEntity> hydrants = world.getEntitiesOfType(StandardEntityURN.HYDRANT);
		if (canadidate < hydrants.size()) {
			EntityID id = world.getUniform().toID(StandardEntityURN.HYDRANT, canadidate);
			Hydrant assignedHydrant = (Hydrant) world.getEntity(id);
			if (!router.isSureReachable(assignedHydrant) &&
					!someoneVisitedArea.contains(assignedHydrant.getID())) {
				move(assignedHydrant, router.getPfCostFunction());
			}
		}
	}

	/**
	 * 近くのHydrantがReachableでなく，そのHydrantに一番近いなら，
	 * 割り当てられて無くても開ける．
	 * @throws ActionCommandException 
	 */
	protected void clearNearHydrant() throws ActionCommandException {
		Collection<StandardEntity> hydrants = world.getEntitiesOfType(StandardEntityURN.HYDRANT);
		if (hydrants.isEmpty()) return;
		Hydrant hydrant = (Hydrant) Collections.min(hydrants, new DistanceSorter(me(), world));
		if (router.isSureReachable(hydrant)) return;

		PoliceForce nearest = me();
		int nearestDist = world.getDistance(hydrant, nearest);
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
			if (se.equals(nearest)) continue;
			PoliceForce pf = (PoliceForce) se;
			if (pf.isXDefined() && pf.isYDefined()) {
				int dist = world.getDistance(hydrant, pf);
				if (dist < nearestDist) {
					nearestDist = dist;
					nearest = pf;
				}
			}
		}
		if (me().equals(nearest)) {
			move(hydrant, router.getPfCostFunction());
		}

	}

	/**
	 * 避難所に向かわせる
	 * @throws ActionCommandException
	 */
	protected void clearRefuge() throws ActionCommandException {
		int canadidate = getUniform() / 2;
		Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		if (canadidate < refuges.size()) {
			EntityID id = world.getUniform().toID(StandardEntityURN.REFUGE, canadidate);
			Refuge assignedRefuge = (Refuge) world.getEntity(id);
			if(!someoneVisitedArea.contains(assignedRefuge.getID())){
				move(assignedRefuge, router.getPfCostFunction());
			}
		}
	}

	/**
	 * 一番近いRefugeがReachableでなく，そのRefugeに一番近いなら，
	 * 割り当てられて無くても開ける．
	 * @throws ActionCommandException 
	 */
	protected void clearNearRefuge() throws ActionCommandException {
		Collection<StandardEntity> refuges = world.getEntitiesOfType(StandardEntityURN.REFUGE);
		if (refuges.isEmpty()) return;
		Refuge refuge = (Refuge) Collections.min(refuges, new DistanceSorter(me(), world));
		if(someoneVisitedArea.contains(refuge.getID())) return;

		PoliceForce nearest = me();
		int nearestDist = world.getDistance(refuge, nearest);
		for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
			if (se.equals(nearest)) continue;
			PoliceForce pf = (PoliceForce) se;
			if (pf.isXDefined() && pf.isYDefined()) {
				int dist = world.getDistance(refuge, pf);
				if (dist < nearestDist) {
					nearestDist = dist;
					nearest = pf;
				}
			}
		}
		if (me().equals(nearest)) {
			move(refuge, router.getPfCostFunction());
		}
	}

	/**
	 * BurningBuildingの近くに移動し、瓦礫を処理
	 * @throws rione.agent.Agent.ActionCommandException
	 */
	protected void clearAroundBurningBuilding() throws ActionCommandException{
		int min = Integer.MAX_VALUE;
		Building temp = null;
		for(Building building : world.getBurningBuildings()){
			if(building.getFierynessEnum() == StandardEntityConstants.Fieryness.INFERNO) continue;
			int distance = world.getDistance(world.getEntity(building.getID()), location());
			if(!router.isSureReachable(building.getID()) && distance < min/* && distance < getDistAverage() * 10*/){
				min = distance;
				temp = building;
			}
		}
		Collection<StandardEntity> aroundEntity = new HashSet<StandardEntity>();
		for(Building building : world.getBurningBuildings()){
			if(building.getFierynessEnum() == StandardEntityConstants.Fieryness.INFERNO) continue;
			Collection<EntityID> r = new HashSet<EntityID>(building.getNeighbours());
			
			for(EntityID e : r){
				if(world.getEntity(e) instanceof Road){
					Collection<Building> neightbourBuilding = new HashSet<Building>();
					neightbourBuilding = getNeighbourBuilding((Road)world.getEntity(e), 10);
					if(neightbourBuilding == null) continue;
					aroundEntity.addAll(neightbourBuilding);
				}
			}
		}
		Building minNeihjbourBuilding = null;
		min = Integer.MAX_VALUE;
		for(StandardEntity se : aroundEntity){
			Building building = (Building)se;
			int distance = world.getDistance(world.getEntity(building.getID()), location());
			if(!router.isSureReachable(building.getID()) && distance < min){
				min = distance;
				minNeihjbourBuilding = building;
			}
		}
		if(minNeihjbourBuilding != null){
			moveFront(minNeihjbourBuilding,router.getPfCostFunction());
		}
		else if(temp!=null){
			moveFront(temp,router.getPfCostFunction());
		}
	}

	protected void clearPartedArea() throws ActionCommandException {
		ArrayList<PartedArea> unreachablePartedArea = new ArrayList<PartedArea>(world.getAreaPartitions());
		for (Iterator<PartedArea> it = unreachablePartedArea.iterator(); it.hasNext();) {
			PartedArea partedArea = it.next();
			if (partedArea.size() < 4 || isReachable(partedArea)) {
				it.remove();
			}
		}
		if (!unreachablePartedArea.isEmpty()) {
			HashSet<Area> dist = new HashSet<Area>(unreachablePartedArea.size() * 4);
			for (PartedArea partedArea : unreachablePartedArea) {
				for (EntityID id : partedArea.getContentBuildingsID()) {
					dist.add((Area) world.getEntity(id));
				}
			}
			moveFront(dist, router.getPfCostFunction());
		}

	}

	private boolean isReachable(PartedArea partedArea) {
		for (EntityID id : partedArea.getContentBuildingsID()) {
			if (router.isSureReachable(id)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 火がついていそうなBuildingに向かう
	 * 
	 * @throws ActionCommandException
	 */
	protected void searchFires() throws ActionCommandException {
		List<Building> ranking = world.getEnergyFlow().getInTotalRanking();
		int u = getUniform();
		if (u < ranking.size()) {
			Building myBuilding = ranking.get(u);
			int lastChanged = world.getTimestamp().getLastChangedTime(myBuilding.getID());
			if (myBuilding.getTotalArea() < world.getEnergyFlow().getInTotal(myBuilding)
					// 後半はあまり見なくて良いと思う
					&& (lastChanged == -1 || time / 3 < time - lastChanged)) {
				moveFront(myBuilding, router.getPfCostFunction());
			}
		}
	}

	protected EntityID Rescuetarget = null;

	protected void moveToLocateInBlockade() throws ActionCommandException {
		if (Rescuetarget == null) {
			ArrayList<DistanceElement> pairs = new ArrayList<DistanceElement>();
			for (Human hm : world.getLocateInBlockade().getLocateInBlocades()) {
				if (hm instanceof PoliceForce)
					continue;
				if (hm instanceof Civilian)
					continue;
				if (hm.isPositionDefined() && hm.isXDefined() && hm.isYDefined()) {
					for (StandardEntity se : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
						PoliceForce police = (PoliceForce) se;
						if (police.isXDefined() && police.isYDefined()) {
							pairs.add(new DistanceElement(police.getID(), hm.getID()));
						}
					}
				}
			}
			Collections.sort(pairs);
			HashSet<EntityID> used = new HashSet<EntityID>();
			for (DistanceElement e : pairs) {
				// TODO DynamicConstants
				if (e.getDistance() > 90000 && getUniform() % 5 < 4)
					break;
				if (used.contains(e.getId1())) {
					continue;
				}
				if (e.getId1().equals(getID())) {
					Rescuetarget = e.getId2();
					move(e.getId2(), router.getPfCostFunction());
				}
				used.add(e.getId1());
			}
		}
		else {
			if (!world.getLocateInBlockade().getLocateInBlocades().contains(world.getEntity(Rescuetarget))) {
				Rescuetarget = null;
			}
			else {
				move(Rescuetarget, router.getPfCostFunction());
			}
		}
	}

	private class DistanceElement implements Comparable<DistanceElement> {
		final private int distance;
		final private EntityID id2;
		final private EntityID id1;

		public DistanceElement(EntityID id1, EntityID id2) {
			this.id1 = id1;
			this.id2 = id2;
			distance = world.getDistance(id1, id2);
		}

		@Override
		public int compareTo(DistanceElement o) {
			return o.getDistance() - getDistance();
		}

		public EntityID getId1() {
			return id1;
		}

		public EntityID getId2() {
			return id2;
		}

		public int getDistance() {
			return distance;
		}
	}

	/**
	 * 除去可能範囲内に埋まっているHumanが居るなら助ける
	 * @throws ActionCommandException
	 */
	protected void clearNearLocateInBlockade() throws ActionCommandException {
		for (Human hm : world.getLocateInBlockade().getLocateInBlocades()) {
			if (hm.isPositionDefined()) {
				Blockade b = world.getLocateInBlockade().getStuck(hm);
				if (b != null && isClearable(me(), b)) {
					expandClear(hm.getX(),hm.getY());
				}
			}
		}
	}


	/**
	 * 全てのUnreachableを開けに行く
	 */
	protected void eraseUnreachable() throws ActionCommandException {
		Set<Area> unreachables = new HashSet<Area>();
		for (StandardEntity se : world.getEntitiesOfType(AgentConstants.AREAS)) {
			Area area = (Area) se;
			if (!router.isSureReachable(area) && !someoneVisitedArea.contains(area.getID())) {
				unreachables.add(area);
			}
		}
		if (unreachables.size() >= 1) {
			move(unreachables, router.getPfCostFunction());
		}
	}

	/**
	 * 救助すべきHumanをreachableにする
	 * 
	 * @throws ActionCommandException
	 */
	protected void clearBuriedHuman() throws ActionCommandException {
		HashSet<Area> humanArea = new HashSet<Area>();
		for (EntityID id : world.getBuriedHumans()) {
			Human hm = (Human) world.getEntity(id);
			if (hm.isPositionDefined()) {
				StandardEntity pos = hm.getPosition(world);
				if (pos instanceof Area) {
					Area area = (Area) pos;
					int dist = world.getDistance(location(), area);
					if (!router.isSureReachable(area)){
						if(someoneVisitedArea.contains(area.getID()) && dist < getDistAverage() * 10){
							continue;
						}
						humanArea.add(area);
					}
				}
			}
		}
		if (!humanArea.isEmpty()) {
			if (humanArea.contains(location())) {
				clearAround();
			}
			move(humanArea, router.getPfCostFunction());
		}
	}

	/**
	 * 自分のいるエリア内で移動せずに瓦礫除去可能な範囲を除去する．<br>
	 * 自分のいる位置から移動せずに隣接エリア内に瓦礫が除去できる場合も除去する．
	 */
	protected void clearAround() throws ActionCommandException {
		ArrayList<Blockade> blocks = new ArrayList<Blockade>();
		for (EntityID id : getVisibleEntities()) {
			StandardEntity entity = world.getEntity(id);
			if (entity instanceof Blockade) {
				Blockade block = (Blockade) entity;
				blocks.add(block);
			}
		}
		// 複数のPFで同時にclearしないようにする
		// Collections.sort(blocks, new DistanceSorter(me(), world));
		Collections.shuffle(blocks, random);
		for (Blockade block : blocks) {
			if (!isClearable(me(),block)) {
				continue;
			}
			expandClear(block);
		}
	}

	/**
	 * criticalAreaを割り当てて開ける
	 * 
	 * @throws ActionCommandException
	 */
	protected void clearCriticalArea() throws ActionCommandException {
		final int uniform = getUniform();
		if (uniform < world.getCriticalArea().size()) {
			move(world.getCriticalArea().get(uniform), router.getPfCostFunction());
		}
	}

	/**
	 * 瓦礫で動けないエージェント
	 * 
	 * @throws ActionCommandException
	 */
	protected void helpAgents() throws ActionCommandException {
		Human nearestAgent = null;
		int min = Integer.MAX_VALUE;

		for (StandardEntity se : world.getEntitiesOfType(AgentConstants.PLATOONS)) {
			Human hm = (Human) se;
			if (hm instanceof PoliceForce) {
				continue;
			}
			if (hm.isPositionDefined()) {
				StandardEntity pos = world.getEntity(hm.getPosition());
				if (!(pos instanceof Road) || pos.equals(location())) {
					continue;
				}
				int dist = world.getDistance(location(), pos);
				Area area = (Area)pos;
				// 最初の100サイクルとかだけにしたほうが良い？
				if (!router.isSureReachable(area)  && dist < getDistAverage() * 4
						&& dist < min){
					min = dist;
					nearestAgent = hm;
				}
			}
		}
		if (nearestAgent != null) {
			move(nearestAgent, router.getPfCostFunction());
		}
	}

	/**
	 * moveしようとしているにも関わらず移動できていないとき
	 * 
	 * @throws ActionCommandException
	 */
	protected void cantMove() throws ActionCommandException {
		MoveToError();
		// 建物から出られないときの対策
		if(getCommandHistory(time-1) != null && getCommandHistory(time-2) != null){
			if (getCommandHistory(time - 1).equals(StandardMessageURN.AK_MOVE)
					&& getCommandHistory(time - 2).equals(StandardMessageURN.AK_MOVE)){
				if (locationHistory.get(time - 1).equals(locationHistory.get(time - 2))) {
					if (time % 4 == 0 && location() instanceof Building) {
						Building bld = (Building) location();
						if(time % 3 == 0){
							errorClear();
						}
						goOut(bld);
					}
					if(time % 3 == 0 && location() instanceof Road){
						if(time % 2 == 0){
							errorClear();
						}
						randomWalk();
					}
				}
			}
		}
	}
	
	/**
	 * BuriedHumanまでをreachableにする（範囲で限定）
	 * PFの位置からすぐにほりに行ける距離である場合のみ作動
	 * 
	 * @throws ActionCommandException
	 */
	protected void cleanBuriedEntrance() throws rione.agent.Agent.ActionCommandException{
		int min = Integer.MAX_VALUE;
		Building nearestBuilding = null;
		Human nearestHuman = null;
		//距離の範囲指定
		if(!(location() instanceof Road))return;
		Collection<Building> buil= getNeighbourBuilding((Road)location(), 6);
		buil.retainAll(world.getObjectsInRange(me(), 5000));
		if(buil.isEmpty())return;
		
		//一番近い人の埋まっているかつ到達不可能な建物を検索
		for(EntityID entity : world.getBuriedHumans()){
			Human buriedHm = (Human) world.getEntity(entity);
			
			Building civBuilding = (Building)world.getEntity(buriedHm.getPosition());
			
			if(!(location() instanceof Road))continue;
			if(buil.contains(civBuilding)) {
				if(!router.isSureReachable((Area)world.getEntity(civBuilding.getID()))
						&& unenteredBuildings.contains(civBuilding.getID())){
					int distance = world.getDistance(civBuilding, location());
					if(!router.isSureReachable(civBuilding) && distance < min){
						min = distance;
						nearestBuilding = civBuilding;
						nearestHuman = buriedHm;
					}
				}
			}
		}
		if(location().equals(nearestBuilding))//一度入った建物に何回も入りそうで怖いんで一応削除を行なっている
			unenteredBuildings.remove(nearestBuilding);
		else if(nearestBuilding != null){//該当した建物に向かう
			move(nearestHuman, router.getPfCostFunction());
		}
	}


	private void errorClear() throws ActionCommandException {
		ArrayList<Blockade> blocks = new ArrayList<Blockade>();
		for (EntityID id : getVisibleEntities()) {
			StandardEntity se = world.getEntity(id);
			if (se instanceof Blockade) {
				Blockade block = (Blockade) se;
				blocks.add(block);
			}
		}
		Collections.sort(blocks, new DistanceSorter(me(), world));
		for (Blockade block : blocks) {
			if (isClearable(me(),block)) {
				expandClear(block);
			}
		}
	}
	
	/**
	 * 高温の建物の周りを探索<br>
	 * 燃えてる建物を見つけやすくする
	 * @throws ActionCommandException
	 * @return 高温建物周辺にしばらく見てない建物があるが、到達不可ならtrue、それ以外ならfalse
	 */
	protected void searchAroundHotBuilding()throws ActionCommandException{
		Building hotBuilding;
		hotBuilding=getHotBuilding();
		if(hotBuilding==null)return;
		
		int minTime=Integer.MAX_VALUE;
		int minDistance=Integer.MAX_VALUE;
		Building DoubtfulBuilding=null;
		
		//温度の高い建物から近い順に建物を6つ取ってくる
		Collection<StandardEntity> AroundHotBuildings = new LinkedList<StandardEntity>(getBuidingExtent(hotBuilding, 6));
		//周りに燃えてる建物があったならそこで終了
		if(AroundHotBuildings.removeAll(world.getBurningBuildings()))return;
		
		
		//TODO rangeの値は検討の余地あり
		//最も前に見た建物に向かう
		for(StandardEntity se : AroundHotBuildings){
			if(se instanceof Building 
					&& !(se instanceof GasStation)
					&& !(se instanceof Refuge)
					&& !(se instanceof FireStation)
					&& !(se instanceof PoliceOffice)
					&& !(se instanceof AmbulanceCentre)) {
				//TODO costfunctionもしくは隣接するRoadなどの値から見ることが困難かを見よう
				int LastSeenTime = world.getTimestamp().getLastSeenTime(se.getID());
				if((!world.getTimestamp().haveSeen(se.getID()) || 20 < time - LastSeenTime)
						&& (LastSeenTime < minTime || (LastSeenTime==minTime && world.getDistance(hotBuilding, se)<minDistance) )){
					DoubtfulBuilding = (Building)se;
					minTime = world.getTimestamp().getLastSeenTime(se.getID());
					minDistance = world.getDistance(hotBuilding, se);
				}
			}
		}
		
		if(DoubtfulBuilding!=null){
			move(DoubtfulBuilding);
		}
	}

	@Override
	public String toString() {
		return "Ri-one police force agent";
	}
}
