package rione.io;

/**
 * byte配列に格納されたbit情報を読み出すクラス
 * @author com
 */
public class BitArrayInputStream {

	/** 出力情報を一時的に格納するバッファ */
	private long buf;
	/** バッファに格納されている情報のサイズ */
	private int offset;
	/** ビット列の格納される配列 */
	private byte[] data;
	/** ビット列の長さ(バイト数) */
	private int length;
	/** ビット列をどこまで読み込んだか */
	private int index;
	
	/**
	 * コンストラクタ
	 * @param data bit情報の格納されたbyte配列
	 */
	public BitArrayInputStream(byte data[]) {
		reset(data);
	}
	
	/**
	 * 最設定します
	 * @param data bit情報の格納されたbyte配列
	 */
	public void reset(byte data[]) {
		this.buf = 0;
		this.offset = 0;
		this.data = data;
		this.length = data.length;
		this.index = 0;
	}

	/**
	 * nビット分の情報を読み込み、intとして返す
	 * @param n 読み込むbitの数
	 * @return 読み込まれた情報をintに格納したもの
	 */
	public int readBit(int n) {
		if (n == 0) {
			return -1;
		}
		while (offset <= n) {
			buf <<= 8;
			buf += read() & 0xFF;
			offset += 8;
		}
		offset -= n;
		int val = (int)(buf >>> offset);
		buf &= 0xFFFFFFFFFFFFFFFFL >>> 64-offset;
		return val;
	}

	/**
	 * 要素を1バイト読み込む
	 * @return 配列の要素1つ分のbit列
	 */
	private int read() {
		if (index >= length) {
			return 0;
		}
		return data[index++];
	}

	/**
	 * 読み込むことのできる要素があるか
	 * @return
	 */
	public boolean hasNext() {
		return index < length || buf != 0;
	}
}