package rione.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rione.communication.Port;
import rione.communication.CommunicationUtil;
import rione.communication.MessageBitSection;
import rione.io.BitArrayInputStream;
import rione.model.route.pov.POVRouter;
import rione.util.BitUtil;

/**
 * ReachableにするべきAreaの集合
 * @author utisam
 *
 */
public class CriticalArea {

	private AdvancedWorldModel world;
	
	ArrayList<StandardEntity> criticalAreas = new ArrayList<StandardEntity>();

	public CriticalArea(AdvancedWorldModel world) {
		this.world = world;
		
		criticalAreas.addAll(world.getEntitiesOfType(StandardEntityURN.REFUGE));
		// Test(3), Kobe(3, 4), berlin(3, 4), Istanble(3, 4), paris(3, 4), VC(2, 3)
		AREAMARK: for (StandardEntity se : world.getEntitiesOfType(AgentConstants.AREAS)) {
			Area area = (Area) se;
			List<EntityID> neighbours = area.getNeighbours();
			if (neighbours.size() < 3)
				continue;
			for (EntityID nid : neighbours) {
				Area neighbour = (Area) world.getEntity(nid);
				if (neighbour instanceof Building)
					continue AREAMARK;
				for (EntityID nnid : neighbour.getNeighbours()) {
					if (world.getEntity(nnid) instanceof Building)
						continue AREAMARK;
				}
			}
			
			
			StandardEntity re = null;
			
			//一番近いREFUGEをとってくる
			for(StandardEntity ref:world.getEntitiesOfType(StandardEntityURN.REFUGE)){
				
				int di = world.getDistance(ref.getID(), se.getID());
				if(re == null){
					re = ref;
					//continue;
				}
				int mindi = world.getDistance(ref.getID(), re.getID());
				if(di < mindi){
					re = ref;
				}
				//避難所からxm以内のところは無視(逆もできるけど、複数避難所があった場合ほぼ全部無視してしまう)
				int x = 50;
				if (world.getDistance(re.getID(), se.getID()) <= x*1000 /*world.getBounds().getHeight()*0.3*/){
					continue AREAMARK;
				}
			}

			
			criticalAreas.add(area);
		}
		// TODO criticalArea評価方法
		// TODO criticalArea生成パターンを増やす
	}
	
	private final int MAX_SEND_SIZE = 7; 
	private int SEND_SIZE_BIT = BitUtil.needBitSize(MAX_SEND_SIZE);
	private List<Area> sendRemovedAreas = new ArrayList<Area>();

	public void update(POVRouter router) {
		sendRemovedAreas.clear();
		for (Iterator<StandardEntity> it = criticalAreas.iterator(); it.hasNext();) {
			Area area = (Area) it.next();
			if (router.isSureReachable(area)) {
				if (sendRemovedAreas.size() <= MAX_SEND_SIZE) {
					sendRemovedAreas.add(area);
				}
				it.remove();
			}
		}
	}

	public Port createPort(final CommunicationUtil comUtil) {
		return new Port() {
			private boolean sendYet = true;
			@Override
			public void init(ChangeSet changed) {
				sendYet = sendRemovedAreas.isEmpty();
			}
			@Override
			public boolean hasNext() {
				return sendYet;
			}
			@Override
			public MessageBitSection next() {
				sendYet = false;
				MessageBitSection sec = new MessageBitSection(10);
				final int n = Math.min(sendRemovedAreas.size(), MAX_SEND_SIZE);
				sec.add(n, SEND_SIZE_BIT);
				for (int i = 0; i < n; i++) {
					comUtil.writeArea(sec, sendRemovedAreas.get(i), world.getUniform());
				}
				return sec;
			}

			@Override
			public void read(EntityID sender, final int time, BitArrayInputStream stream) {
				// ビット化け対策
				final int n = Math.max(0, stream.readBit(BitUtil.needBitSize(MAX_SEND_SIZE)));
				for (int i = 0; i < n ; i++) {
					EntityID id = comUtil.readArea(stream, world.getUniform());
					if (criticalAreas.remove(world.getEntity(id))) {
						//System.out.println("revive communication remove");
					}
				}
			}
		};
	}

	public int size() {
		return criticalAreas.size();
	}

	public Area get(int index) {
		return (Area) criticalAreas.get(index);
	}

	public List<StandardEntity> getAreas() {
		return criticalAreas;
	}

}
