package rione.model;

import java.util.HashSet;
import java.util.List;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

final public class IsolatedArea extends HashSet<StandardEntity> {
	private static final long serialVersionUID = 1L;
	
	private AdvancedWorldModel world = null;
	
	public IsolatedArea(StandardEntity me, AdvancedWorldModel w) {
		super(w.getEntitiesOfType(AgentConstants.AREAS));
		world = w;
		Area area;
		if (me instanceof Human) {
			area = (Area) ((Human) me).getPosition(world);
		}
		else {
			area = (Building) me;
		}
		setIsolationEntities(area);
	}

	private void setIsolationEntities(Area area) {
		if (!this.contains(area)) {
			return;
		}
		this.remove(area);
		List<EntityID> neighbors = area.getNeighbours();
		for (EntityID id : neighbors) {
			Area a = (Area) world.getEntity(id);
			setIsolationEntities(a);
		}		
	}
	
	/**
	 * 再帰的につながっているAreaを孤立しているAreaのSetから除外する
	 */
	public boolean isIsolation(EntityID position) {
		try {
			return isIsolation((Area) world.getEntity(position));
		}catch (ClassCastException e) {
			return false;
		}
	}

	/**
	 * 孤立しており、行くことができない場所であるかどうか。
	 * 
	 * @param area
	 *            判定するArea
	 * @return 行くことができないならばtrue
	 */
	public boolean isIsolation(Area area) {
		return this.contains(area);
	}	
}
