package rione.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import rione.util.IdSorter;

final public class Uniform {
	private Map<EntityID, Integer> idToUniform = new HashMap<EntityID, Integer>(40);
	private Map<StandardEntityURN, Map<Integer, EntityID>> uniformToID = new HashMap<StandardEntityURN, Map<Integer, EntityID>>(40);

	public Uniform(AdvancedWorldModel world) {
		for (StandardEntityURN urn : AgentConstants.UNIFORMABLE_ENTITIES) {
			Map<Integer, EntityID> uniform_id = new HashMap<Integer, EntityID>();
			ArrayList<StandardEntity> entities = new ArrayList<StandardEntity>(world.getEntitiesOfType(urn));
			// EntityID順にソート
			Collections.sort(entities, new IdSorter());
			int i = 0;
			for (StandardEntity entity : entities) {
				idToUniform.put(entity.getID(), i);
				uniform_id.put(i, entity.getID());
				i++;
			}
			uniformToID.put(urn, uniform_id);
		}
	}

	public int toUniform(EntityID id) {
		Integer uniform = idToUniform.get(id);
		if (uniform == null) {
			return -1;
		}
		return uniform.intValue();
	}

	public EntityID toID(StandardEntityURN urn, int uniform) {
		if (!AgentConstants.UNIFORMABLE_ENTITIES.contains(urn)) {
			return null;
		}
		return uniformToID.get(urn).get(uniform);
	}
}
