package rione.model.route.area;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Road;

import rione.model.AdvancedWorldModel;

public abstract class CostFunction {

	public abstract int cost(Area from, Area through, Area to);
	public abstract int cost(Area from, Area to);


	/**
	 * デフォルトのコストファンクション
	 */
	public static CostFunction normal(final ReachableArea reachable, final AdvancedWorldModel world) {
		return new CostFunction() {
			@Override
			public int cost(Area from, Area through, Area to) {
				Edge e1 = from.getEdgeTo(through.getID());
				Edge e2 = to.getEdgeTo(through.getID());
				int sx = (e1.getStartX() + e1.getEndX()) / 2;
				int sy = (e1.getStartY() + e1.getEndY()) / 2;
				int ex = (e2.getStartX() + e2.getEndX()) / 2;
				int ey = (e2.getStartY() + e2.getEndY()) / 2;
				int c = (int) (Math.hypot(sx-ex, sy-ey));
//				//c = 1000000/c;
//				int c = 50;
				
				Edge edge = through.getEdgeTo(to.getID());

				double roadWidth = Math.hypot(edge.getStartX()-edge.getEndX(), edge.getStartY()-edge.getEndY());
				if (roadWidth<2000) {
					c += 100000;
				}
//				else if (roadWidth > 3000) {
//					c /= 10;
//				}
				// if (through instanceof Road && !reachable.isPassable((Road)through)) {
				if (through instanceof Road && from instanceof Road && to instanceof Road && !reachable.isPassable((Road)from, (Road)through, (Road)to)) {
					c += 100000;
				}
 				else if (through instanceof Building) {
					c += 20000;
				}
				return c;
			}

			@Override
			public int cost(Area from, Area to) {
				final int WEIGHT = 10; 
				int c = 1;
				Edge edge = from.getEdgeTo(to.getID());
				if (Math.hypot(edge.getStartX()-edge.getEndX(), edge.getStartY()-edge.getEndY())<2000) {
					c *= WEIGHT * WEIGHT;
				}
				if (to instanceof Road && !reachable.isPassable((Road)to)) {
					//System.out.println(world.me().getID()+"\t"+from.getID()+" from "+to.getID());
					c *= WEIGHT * WEIGHT * WEIGHT * WEIGHT;
				}
				if (to instanceof Building) {
					c *= WEIGHT;
				}
				return c;
			}
		};
	}

	public static CostFunction pf(final ReachableArea reachable, final AdvancedWorldModel world) {
		return new CostFunction() {
			@Override
			public int cost(Area from, Area through, Area to) {
				final int WEIGHT = 10; 
				//int c = 10;
//				Edge e1 = from.getEdgeTo(through.getID());
//				Edge e2 = to.getEdgeTo(through.getID());
//				int sx = (e1.getStartX() + e1.getEndX()) / 2;
//				int sy = (e1.getStartY() + e1.getEndY()) / 2;
//				int ex = (e2.getStartX() + e2.getEndX()) / 2;
//				int ey = (e2.getStartY() + e2.getEndY()) / 2;
//				int c = (int) (Math.hypot(sx-ex, sy-ey)/10);
				int c = 10;
				//System.out.println(c);
				Edge edge = through.getEdgeTo(to.getID());
				double roadWidth = Math.hypot(edge.getStartX()-edge.getEndX(), edge.getStartY()-edge.getEndY());
				
				if (roadWidth<2000) {
					c *= WEIGHT * WEIGHT * WEIGHT * WEIGHT;
				}
				else if (roadWidth > 3000) {
					c /= 10;
				}
				if (through instanceof Building) {
					c *= WEIGHT * WEIGHT * WEIGHT * WEIGHT;
				}
				return c;
			}

			@Override
			public int cost(Area from, Area to) {
				final int WEIGHT = 10; 
				int c = 1;
				Edge edge = from.getEdgeTo(to.getID());
				if (Math.hypot(edge.getStartX()-edge.getEndX(), edge.getStartY()-edge.getEndY())<2000) {
					c *= WEIGHT * WEIGHT * WEIGHT * WEIGHT;
				}
				if (to instanceof Building) {
					c *= WEIGHT * WEIGHT * WEIGHT * WEIGHT;
				}
				return c;
			}
		};
	}


}
