package rione.test;

import static org.junit.Assert.*;

import java.util.Arrays;

import org.junit.Test;

import rione.util.UnionFindTree;

public class UnionFindTreeTest {

	@Test(expected = NullPointerException.class)
	public void testNullAdd() {
		UnionFindTree<Integer> tree = new UnionFindTree<Integer>();
		tree.add(null);
	}

	@Test
	public void testAddAll() {
		UnionFindTree<Integer> tree = new UnionFindTree<Integer>();
		assertTrue(tree.addAll(Arrays.asList(new Integer[] {1, 2, 3})));
		assertFalse(tree.addAll(Arrays.asList(new Integer[] {1, 2, 3})));
		
		assertTrue(tree.same(1, 1));
		assertTrue(tree.same(2, 2));
		assertFalse(tree.same(1, 2));
	}

	@Test
	public void testUnite() {
		UnionFindTree<Integer> tree = new UnionFindTree<Integer>();
		assertTrue(tree.add(1));
		assertTrue(tree.add(2));
		assertFalse(tree.add(2));
		
		assertTrue(tree.same(1, 1));
		assertTrue(tree.same(2, 2));
		assertFalse(tree.same(1, 2));
		
		tree.unite(1, 2);
		
		assertTrue(tree.same(1, 1));
		assertTrue(tree.same(2, 2));
		assertTrue(tree.same(1, 2));
	}

	@Test
	public void testClear() {
		UnionFindTree<Integer> tree = new UnionFindTree<Integer>();
		tree.add(1);
		tree.add(2);
		assertEquals(tree.size(), 2);
		tree.clear();
		assertEquals(tree.size(), 0);
		assertTrue(tree.isEmpty());
	}

	@Test
	public void testIterator() {
		UnionFindTree<Integer> tree = new UnionFindTree<Integer>();
		Integer[] testCase = new Integer[] {1, 2};
		tree.addAll(Arrays.asList(testCase));
		int i = 0;
		for (Integer x : tree) {
			assertEquals(testCase[i], x);
			i++;
		}
	}

	@Test
	public void testRemove() {
		UnionFindTree<Integer> tree = new UnionFindTree<Integer>();
		tree.addAll(Arrays.asList(new Integer[] {1, 2}));
		tree.unite(1, 2);
		assertTrue(tree.same(1, 2));
		tree.remove(1);
		assertFalse(tree.contains(1));
		assertTrue(tree.contains(2));
		
		tree.clear();
		tree.addAll(Arrays.asList(new Integer[] {1, 2, 3}));
		tree.unite(1, 2);
		tree.unite(1, 3);
		assertTrue(tree.remove(2));
		assertEquals(tree.size(), 2);
	}

	@Test
	public void testRetainAll() {
		UnionFindTree<Integer> tree = new UnionFindTree<Integer>();
		tree.addAll(Arrays.asList(new Integer[] {1, 2, 3, 4}));
		tree.retainAll(Arrays.asList(new Integer[] {1}));
		assertEquals(tree.size(), 1);
	}

}
